#!/usr/bin/env python3
# Created by xiazeng on 2019-05-22

import logging
import time
import os
import shutil
import json
import threading
import types
import typing
from functools import wraps

from minium.utils.utils import timeout
from ..exception import *
from ...utils.emitter import ee
from .wording import Language, WORDING

logger = logging.getLogger("minium")


def call_funtion_log(func, name: str):
    @wraps(func)
    def wrapper(*args, **kwargs):
        logger.info(f"call {name}")
        result = func(*args, **kwargs)
        if name.find(".handle") > 0 or name.find(".allow") > 0:
            logger.info(f"{name} return {str(result)}")
        logger.info(f"call {name} end")
        return result

    return wrapper


class NativeMetaClass(type):
    def __new__(mcs, cls_name, bases, attr_dict):
        for k, v in attr_dict.items():
            if isinstance(v, types.FunctionType) and not k.startswith("_"):
                attr_dict[k] = call_funtion_log(v, f"{cls_name}.{k}")
        return type.__new__(mcs, cls_name, bases, attr_dict)


class BaseNative(object, metaclass=NativeMetaClass):
    _require_conf_: typing.List[typing.Union[str, tuple, list]] = []  # 配置必要的字段
    def __init__(self, json_conf):
        self._check_config(json_conf or {})  # 检查必要的配置字段
        self.json_conf = json_conf
        self.perf_data = []
        self.__auto_authorize_thread = None
        self.__auto_authorize_lock = threading.Condition()
        self.__auto_authorize_flag = False
        self.__last_notify_time = time.time()
        self.auto_authorize = False  # 是否自动授权
        self.language = Language.zh  # 默认中文, start wechat时检测

    def _check_config(self, conf):
        """检查配置必要字段

        :param dict conf: 配置dict
        """
        for key in self._require_conf_:
            if isinstance(key, (tuple, list)):
                conform = False
                for k in key:
                    if k in conf:
                        conform = True
                        break
                if not conform:
                    raise KeyError(f"({' or '.join(key)}) is required in instantiate {self.__class__.__name__}")
            elif key not in conf:
                raise KeyError(f"{key} is required in instantiate {self.__class__.__name__}")

    def release(self):
        self.release_auto_authorize()

    def start_wechat(self):
        """
        启动微信
        :return:
        """
        raise NotImplementedError()

    def stop_wechat(self):
        """
        启动微信
        :return:
        """
        raise NotImplementedError()

    def connect_weapp(self, path):
        """
        扫码图片
        :param path:图片名称
        :return:
        """
        raise NotImplementedError()

    def screen_shot(self, filename, return_format="raw"):
        """
        截图
        :param filename: 文件存放的路径
        :param return_format: 除了将截图保存在本地之外, 需要返回的图片内容格式: raw(default) or pillow
        :return: raw data or PIL.Image
        """
        raise NotImplementedError()

    def pick_media_file(
        self,
        cap_type="camera",
        media_type="photo",
        original=False,
        duration=5.0,
        names=None,
    ):
        """
        获取媒体文件
        :param cap_type: camera: 拍摄 | album: 从相册获取
        :param names: 传入一个 list 选择多张照片或者视频(照片和视频不能同时选择, 且图片最多 9 张, 视频一次只能选择一个)
        :param media_type: photo 或者 video
        :param duration: 拍摄时长
        :param original: 是否选择原图(仅图片)
        :return:
        """
        raise NotImplementedError()

    def input_text(self, text):
        """
        input 组件填写文字
        :param text: 内容
        :return:
        """
        raise NotImplementedError()

    def input_clear(self):
        """
        input 组件清除文字(使用此函数之前必须确保输入框处于输入状态)
        :return:
        """
        raise NotImplementedError()

    def textarea_text(self, text, index=1):
        """
        给 textarea 输入文字需要在小程序层提供该 textarea 的 index 信息
        :param text: 内容
        :param index: 多个 textarea 同时存在一个页面从上往下排序, 计数从 1 开始
        :return:
        """
        raise NotImplementedError()

    def textarea_clear(self, index=0):
        """
        给 textarea 清除文字需要在小程序层提供该 textarea 的 index 信息
        :param index: 多个 textarea 同时存在一个页面从上往下排序, 计数从 1 开始
        :return:
        """
        raise NotImplementedError()

    def allow_authorize(self, answer=True, title=None):
        """
        处理授权确认弹框
        :param answer: True or False
        :return: bool
        """
        raise NotImplementedError()

    def allow_login(self, answer=True):
        """
        处理微信登陆确认弹框
        :param answer: True or False
        :return:
        """
        raise NotImplementedError()

    def allow_get_user_info(self, answer=True):
        """
        处理获取用户信息确认弹框
        :param answer: True or False
        :return:
        """
        raise NotImplementedError()

    def allow_get_location(self, answer=True):
        """
        处理获取位置信息确认弹框
        :param answer: True or False
        :return:
        """
        raise NotImplementedError()

    def allow_get_we_run_data(self, answer=True):
        """
        处理获取微信运动数据确认弹框
        :param answer: True or False
        :return:
        """
        raise NotImplementedError()

    def allow_record(self, answer=True):
        """
        处理获取录音权限确认弹框
        :param answer: True or False
        :return:
        """
        raise NotImplementedError()

    def allow_write_photos_album(self, answer=True):
        """
        处理保存相册确认弹框
        :param answer: True or False
        :return:
        """
        raise NotImplementedError()

    def allow_camera(self, answer=True):
        """
        处理使用摄像头确认弹框
        :param answer: True or False
        :return:
        """
        raise NotImplementedError()

    def allow_get_user_phone(self, answer=True):
        """
        处理获取用户手机号码确认弹框
        :param answer: True or False
        :return:
        """
        raise NotImplementedError()

    def allow_send_subscribe_message(self, answer=True):
        """
        允许发送订阅消息
        """
        raise NotImplementedError()

    def handle_modal(self, btn_text="确定", title=None, index=-1, force_title=False):
        """
        处理模态弹窗
        :param title: 传入弹窗的 title 可以校验当前弹窗是否为预期弹窗
        :param btn_text: 根据传入的 name 进行点击
        :return:
        """
        raise NotImplementedError()

    def handle_action_sheet(self, item):
        """
        处理上拉菜单
        :param item: 要选择的 item
        :return:
        """
        raise NotImplementedError()

    def forward_miniprogram(
        self, name: str, text: str = None, create_new_chat: bool = True
    ):
        """
        通过右上角更多菜单转发小程序
        :type text: 分享携带的内容
        :param names: 要分享的人
        :param create_new_chat: 是否创建群聊
        :return:
        """
        raise NotImplementedError()

    def forward_miniprogram_inside(
        self, name: str, text: str = None, create_new_chat: bool = True
    ):
        """
        小程序内触发转发小程序
        :param names: 要分享的人
        :param create_new_chat: 是否创建群聊
        :return:
        """
        raise NotImplementedError()

    def send_custom_message(self, message=None):
        """
        处理小程序im 发送自定义消息
        :param message: 消息内容
        :return:
        """
        raise NotImplementedError()

    def phone_call(self):
        """
        处理小程序拨打电话
        :return:
        """
        raise NotImplementedError()

    def map_select_location(self, name=None):
        """
        原生地图组件选择位置
        :param name: 位置名称
        :return:
        """
        raise NotImplementedError()

    def map_back_to_mp(self):
        """
        原生地图组件查看定位,返回小程序
        :return:
        """
        raise NotImplementedError()

    def deactivate(self, duration):
        """
        使微信进入后台一段时间, 再切回前台
        :param duration: float
        :return: NULL
        """
        raise NotImplementedError()

    def get_authorize_settings(self):
        """
        在小程序的授权页面，获取小程序的授权设置
        :return:
        """
        raise NotImplementedError()

    def back_from_authorize_setting(self):
        """
        从小程序授权页面跳转回小程序
        :return:
        """
        raise NotImplementedError()

    def authorize_page_checkbox_enable(self, name, enable):
        """
        在小程序授权设置页面操作CheckBox
        :param name: 设置的名称
        :param enable: 是否打开
        :return:
        """
        raise NotImplementedError()

    def start_get_perf(self, timeinterval=15):
        """
        开始获取性能数据
        :param timeinterval: 抽样时间间隔
        :return: boolen
        """
        return False

    def stop_get_perf(self):
        """
        停止获取性能数据
        :return: string: json.dumps([{cpu, mem, fps, timestamp}])
        """
        pass

    def get_perf_data(
        self, start_timestamp: int = None, end_timestamp: int = None
    ) -> str:
        """
        获取[start_timestamp, end_timestamp]区间内的perf data
        :start_timestamp: 开始时间戳，如果为none，不过滤开始时间，start_timestamp后没有任何数据且perf data列表有数据，返回最后一个数据(理解为数据没有变化)
        :end_timestamp: 结束时间戳，如果为none，不过滤结束时间, 如果end_timestamp比第一个元素时间还小，返回空
        :return: "[{timestamp, cpu, mem, fps}]" or ""
        """
        if not self.perf_data:
            return ""
        if start_timestamp is None and end_timestamp is None:
            return json.dumps(self.perf_data)
        end_index = len(self.perf_data)  # 如果之后还有数据入列表也忽略，数据从后往前搜索
        if end_timestamp is not None:  # search end_index
            while self.perf_data[end_index - 1]["timestamp"] > end_timestamp:
                end_index -= 1
                if end_index == 0:
                    break
        if end_index == 0:
            return ""
        if start_timestamp is None or end_index == 1:
            return json.dumps(self.perf_data[:end_index])
        else:  # search start_index
            start_index = end_index - 1  # at least equal 1
            while self.perf_data[start_index - 1]["timestamp"] >= start_timestamp:
                start_index -= 1
                if start_index == 0:
                    break
            return json.dumps(self.perf_data[start_index:end_index])

    def get_start_up(self):
        """
        获取小程序启动时间
        """
        return 0

    def start_get_start_up(self):
        """
        开始监听启动时间
        """
        pass

    def stop_get_start_up(self):
        """
        结束监听启动时间
        :return: number
        """
        return 0

    def click_coordinate(self, x, y):
        """
        按坐标点击
        :param x:
        :param y:
        :return:
        """
        raise NotImplementedError()

    def get_pay_value(self):
        """
        获取支付金额
        """
        raise NotImplementedError()

    def input_pay_password(self):
        """
        输入支付密码
        """
        raise NotImplementedError()

    def close_payment_dialog(self):
        """
        关闭支付弹窗
        """
        raise NotImplementedError()

    def text_exists(self, text="", iscontain=False, wait_seconds=5):
        """
        检测是否存在text
        """
        raise NotImplementedError()

    def text_click(self, text="", iscontain=False):
        """
        点击内容为text的控件
        """
        raise NotImplementedError()

    def hide_keyboard(self):
        """
        隐藏键盘
        :return:
        """
        raise NotImplementedError()

    def is_app_in_foreground(self, appid):
        """
        判断{appid}的小程序是否在前台，默认不会退后台
        :return: bool
        """
        return True

    def check_connection(self, *args):
        """
        检查真机调试通道是否还存在, 如果断连则重连
        :return: bool
        """
        return True

    def _get_current_activity(self):
        return ""

    def _is_in_wechat(self, activity: str):
        return True

    def _is_in_wechat_main(self, activity: str):
        return False

    def _is_in_miniprogram(self, activity: str):
        """
        需要甄别: 插件页面、webview页面、普通页面
        """
        return True

    def _is_in_target_miniprogram(self, appid: str):
        return True

    def _close_miniprogram(self):
        return True

    def _get_any_modal(self, confirm=False):
        return None

    def _is_in_payment(self):
        return False

    def _handle_modal(self, modal):
        if not modal:
            return ModalStatus.OK
        logger.info(f"handle modal: {modal}")
        if modal.click_if_exists(0.5):
            return ModalStatus.OK
        return ModalStatus.Error

    def _back_to_target_miniprogram(self, appid):
        if not self._is_in_target_miniprogram(appid):
            logger.error(f"not in {appid}")
            # 先判断是不是目标小程序, 尝试3次
            for i in range(1, 4):
                logger.error(f"try to close miniprogram {i} times")
                if not self._close_miniprogram():  # 无法操作关闭小程序
                    logger.error("close miniprogram error, relaunch")
                    return ResetError.RELAUNCH_APP
                time.sleep(1)
                if self._is_in_target_miniprogram(appid):
                    logger.info(f"now in {appid}")
                    break
                if i == 3:  # 3次仍未回到目标小程序
                    logger.error(f"try to close miniprogram {i} times but still error, relaunch")
                    return ResetError.RELAUNCH_APP
        has_any_modal = self._get_any_modal(self.allow_authorize)  # 获取目标小程序中的modal
        # 已经回到当前小程序
        if not has_any_modal:
            return ResetError.OK
        return has_any_modal

    def _handle_any_modal(self, has_any_modal, cnt, crash_cnt):
        ret = None
        while has_any_modal:  # 有弹窗, 点击直到没有
            ret = self._handle_modal(has_any_modal)
            if ret == ModalStatus.OK:
                cnt = 3
                crash_cnt -= 1  # 虽然点击指令执行成功了，但很多情况是没有点击到的
                time.sleep(1)
            elif ret == ModalStatus.NotOfficialModal:
                # 操作成功了，但弹窗可能并不是官方提供的弹窗
                cnt -= 1
                crash_cnt -= 1  # 虽然点击指令执行成功了，但很多情况时没有点击到的
                time.sleep(1)
            elif ret == ModalStatus.NotFound:
                break
            has_any_modal = self._get_any_modal(self.allow_authorize)
            if ret == ModalStatus.Error:
                break
            if crash_cnt < 0:
                break
        return ret, has_any_modal, cnt, crash_cnt

    def back_to_miniprogram(self, appid=""):
        """
        返回小程序
        :return: ResetError, 0: 返回成功. -1: 无法识别的错误. -2: 需要重新加载小程序. -3: 需要重新加载微信
        """
        crash_cnt = 15  # 防止预期之外的死循环, 每一次操作都会计数减一
        cnt = 3  # 防止一直检测循环, 当有检测操作成功的，重置为初始次数
        last_activity = None
        while cnt > 0:
            logger.info(f"try back_to_miniprogram {cnt}")
            cnt -= 1
            # 检测是否还在微信
            current_activity = self._get_current_activity()
            if not self._is_in_wechat(current_activity):
                logger.error("current activity is %s, not wechat" % current_activity)
                return ResetError.RELAUNCH_APP
            # 已经回到了微信主程序
            if self._is_in_wechat_main(current_activity):
                logger.error(
                    "already back to wechat main activity, please relaunch miniprogram"
                )
                return ResetError.RELAUNCH_MINIPROGRAM

            has_any_modal = None
            # 回到目标小程序
            if self._is_in_miniprogram(current_activity):
                ret = self._back_to_target_miniprogram(appid)
                if isinstance(ret, ResetError):
                    return ret
                has_any_modal = ret
            else:
                logger.error("not in miniprogram")
                has_any_modal = self._get_any_modal(self.allow_authorize)  # 不在小程序中也需要尝试检查一下弹窗
            # 看看是不是支付的弹窗
            if self._is_in_payment():
                logger.warning("有支付弹窗")
                if self.close_payment_dialog():
                    logger.info("成功关闭支付弹窗")
                    cnt = 3
                    crash_cnt -= 1
                continue
            # 无论是否在小程序进程，都可能遗留弹窗，如果检测到有，则点击"取消/拒绝"
            ret, has_any_modal, cnt, crash_cnt = self._handle_any_modal(
                has_any_modal, cnt, crash_cnt
            )
            logger.info(f"finish handle modals {ret}")
            if ret is not None:
                current_activity = (
                    self._get_current_activity()
                )  # 操作完可能对当前activity有所影响，应更新
            if not has_any_modal and self._is_in_miniprogram(current_activity):
                logger.info(f"back to miniprogram ok, current activity: {current_activity}")
                return ResetError.OK
            if (
                last_activity and last_activity != current_activity
            ):  # activity有变化了，认为操作有效可以reset循环
                cnt = 3
            last_activity = current_activity
            # 返回
            if not self._is_in_miniprogram(current_activity):
                self._press_back()
                crash_cnt -= 1
                time.sleep(1)
            if crash_cnt < 0:
                break
        logger.error(f"back to miniprogram error, current activity: {self._get_current_activity()}")
        return ResetError.ERROR

    def _press_back(self):
        """
        返回
        :return: bool
        """
        return True

    # @ee.on('notify')  # 不能直接修饰成员函数
    def notify(self):
        """
        通知auto authorize线程处理弹窗. 通知时机包括:
        1. tap/long_press/click之后
        2. after hook auth method
        """
        if not self.__auto_authorize_thread:
            return
        logger.debug("receive a notify signal")
        self.__notify()

    def set_auto_authorize(self, answer=True):
        """
        自动操作授权，操作时机等待 notify
        :params answer: True: 自动同意, False: 自动拒绝
        :return: bool
        """
        self.auto_authorize = answer
        if self.__auto_authorize_thread:
            return True
        self.__auto_authorize_thread = threading.Thread(
            target=self.__op_auto_authorize, args=(answer,)
        )
        self.__auto_authorize_thread.setDaemon(True)
        self.__auto_authorize_thread.start()
        ee.on("notify", self.notify)  # 监听相应函数
        return True

    def release_auto_authorize(self):
        """
        release auto authorize thread
        """
        if self.__auto_authorize_thread:
            # 移除监听事件
            ee.remove_all_listeners("notify")
            self.__auto_authorize_flag = False
            self.__notify()
            logger.info("stopping auto authorize thread")
            self.__auto_authorize_thread.join()
            logger.info("stop auto authorize thread")
            self.__auto_authorize_thread = None

    def close_local_debug_modal(self):
        self.handle_modal(
            btn_text=WORDING.COMMON.MODAL_CONFIRM.value, title=WORDING.COMMON.LOCAL_DEBUG_MODAL_TITLE.value, force_title=True
        )

    # protect method
    def _wait(self, func, timeout, interval=1):
        """
        等待直到`func`为true
        :func: callable function
        :timeout: timeout
        :interval: query step
        :return: bool
        """
        # func = lambda :True or False
        if not callable(func):
            return False
        s = time.time()
        while time.time() - s < timeout:
            if func():
                return True
            time.sleep(interval)
        return False

    def _auto_authorize_callback(self, answer=True):
        """
        处理authorize的逻辑, 可重载
        """
        ret = True
        cnt = 10  # 最大处理10个授权窗口，防止死循环(检测到有元素 & 可以click，但实际点不动)
        while ret and cnt > 0:
            cnt -= 1
            try:
                ret = self.allow_authorize(answer)
            except NotImplementedError:
                #  未实现的处理方法，直接break进程
                return False
            time.sleep(1)
        return cnt > 0

    def _empty_base_screen_dir(self, dirname):
        if os.path.exists(dirname):
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)
                time.sleep(1)
            else:
                os.remove(dirname)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    def __notify(self):
        self.__auto_authorize_lock.acquire()
        self.__auto_authorize_lock.notify_all()
        self.__last_notify_time = time.time() + 6  # 收到通知后6秒内还应该继续检测, 一般弹窗有滞后性
        self.__auto_authorize_lock.release()

    def __op_auto_authorize(self, answer):
        self.__auto_authorize_flag = True
        while self.__auto_authorize_flag:
            self.__auto_authorize_lock.acquire()
            ret = self.__auto_authorize_lock.wait(timeout=5)
            self.__auto_authorize_lock.release()
            if not ret and self.__last_notify_time < time.time():
                continue
            if not self.__auto_authorize_flag:
                break
            logger.debug("before auto_authorize_callback")
            ret = self._auto_authorize_callback(answer)
            logger.debug(
                "after auto_authorize_callback, %s. flag: %s"
                % (ret, self.__auto_authorize_flag)
            )
            if not ret:
                break


class NativeError(RuntimeError):
    pass



region = "en"
try:
    import locale
    language_code = locale.getlocale()[0]
    if language_code is None:
        language_code = locale.getdefaultlocale()[0]
    if language_code:
        region = language_code.split("_")[0]
except ImportError:
    pass


DEFAULT_LOCALE = "en"
if region == "zh":
    CURRENT_LOCALE = "zh"
else:
    CURRENT_LOCALE = "en"


def item(**kwargs):
    if CURRENT_LOCALE in kwargs:
        value = kwargs.get(CURRENT_LOCALE)
    elif DEFAULT_LOCALE in kwargs:
        value = kwargs.get(DEFAULT_LOCALE)
    elif len(kwargs.items()) > 0:
        value = list(kwargs.items())[0][1]
    else:
        value = ""
    return value


class obj(dict):
    def __init__(self, from_dict=None):
        if isinstance(from_dict, dict):
            for k, v in from_dict.items():
                setattr(self, k, v)
        super(obj, self).__init__(self.__dict__)

    def __getattr__(self, __k):
        try:
            return self[__k]
        except KeyError:
            return None

    def __setattr__(self, __k, __v):
        if isinstance(__v, dict):
            self[__k] = obj(__v)
        else:
            self[__k] = __v


class WORDING:
    class HELP:
        version = item(en="version", zh="版本")
        path = item(
            en = "case directory, default current directory",
            zh = "用例目录, 默认当前目录"
        )
        module = item(
            en = "case package name, usually means a python file name",
            zh = "模块名, 一般可以在python中import"
        )
        case_name = item(
            en = "case name, usually means a function",
            zh = "用例名, 一般为MiniTest类的成员函数"
        )
        module_search_path = item(
            en = "you can add some custom module into sys path by this param",
            zh = "指定python依赖的自定义模块路径, 同sys.path.append"
        )
        apk = item(
            en = "show apk path which may you need to install before running test in android device",
            zh = "运行安卓真机自动化时, 需要安装到手机的测试apk"
        )
        suite = item(
            en = "test suite file, a json format file or just a json string",
            zh = "测试套件json文件, 一般指suite.json"
        )
        config = item(
            en = "config file path",
            zh = "配置文件, 一般指config.json"
        )
        generate = item(
            en = "generate html report",
            zh = "生成html测试报告"
        )
        mode = item(
            en = "cases run mode, parallel or fork",
            zh = "运行测试用例的模式, 包括parallel(用例分配到多个账号同时测试)和fork(多个账号都分别运行全部的测试用例)"
        )
        accounts = item(
            en = "get accounts which already login",
            zh = "获取除登录主账号外的测试账号, 详情见开发者工具【工具->多账号调试】"
        )
        only_native = item(
            en = "Only init native driver",
            zh = "仅仅实例化Native驱动, 控制手机客户端"
        )
        test_connection = item(
            en = "test connection between minium and ide",
            zh = "测试minium驱动的websocket链接是否正常"
        )
        test_port = item(
            en = "test connection port, default: 9420",
            zh = "指定自动化测试端口"
        )
        task_limit_time = item(
            en = "task max runtime, default: 0, unlimited",
            zh = "测试任务最大运行时间, 超时将直接终止测试进程. 默认0, 不限时"
        )
        just_test = item(
            en = "just test loader, not run session",
            zh = "测试参数/配置文件等是否正确, 不真正运行用例"
        )
        check_env = item(
            en = "check system environment for minitest",
            zh = "检查minitest的环境"
        )




#!/usr/bin/env python3
# Created by xiazeng on 2020/5/27
import logging
import at.utils.decorator
import at.webdriver.driver as driver
import at.webdriver.tabwebsocket as tabwebsocket


class MiniApp:
    def __init__(self, at_instance, remote_port_name):
        """
        对小程序的封装
        :type at_instance: at.At
        :param remote_port_name:
        """
        self.at = at_instance
        self.port = driver.pick_unuse_port()
        self.remote_port_name = remote_port_name
        self.adb = at_instance.adb
        self.adb.forward_name(f"tcp:{self.port}", f"localabstract:{remote_port_name}")

    def __del__(self):
        self.adb.forward_remove(self.port)

    @at.utils.decorator.retry_in(30, 2)
    def get_current_page(self):
        chrome_tabs = driver.tabs(self.port)
        for tab in chrome_tabs:
            if tab and "title" in tab and tab["title"].endswith(":VISIBLE"):
                logging.info('find port: %s -> %s, with title: %s', self.remote_port_name, self.port, tab['title'])
                if "webSocketDebuggerUrl" not in tab:
                    raise RuntimeError("小程序调试websocket被占用")
                sock = tabwebsocket.TabWebSocket(tab["webSocketDebuggerUrl"])
                # __wxConfig是公共库的全局变量
                lib_version = sock.run_script_with_output("window.__wechatLibVersion__")
                logging.info("lib version:%s" % lib_version)
                return driver.WebPage(sock, self.at)
        return None



# coding:utf-8
import requests


class PrivateKeyAuth(requests.auth.AuthBase):
    """
    基于 Private Key 的认证方式, 适用于开发者用自己的 private key 调接口

    @see: http://x.code.oa.com/Code/Git/blob/master/help/api/README.md#private-key%E6%96%B9%E5%BC%8F
    """

    def __init__(self, private_key):
        self.private_key = private_key

    def __call__(self, r):
        r.headers["PRIVATE-TOKEN"] = self.private_key
        return r


class OAuthToken(requests.auth.AuthBase):
    """
    基于 OAuth 的认证方式, 注: 并不负责生成 OAuth 令牌, 适用于服务商通过用户的 OAuth 令牌来调接口

    @see: http://x.code.oa.com/Code/Git/blob/master/help/solutions/oauth2.md#%E4%BD%BF%E7%94%A8%E8%AE%BF%E9%97%AE%E4%BB%A4%E7%89%8C%E8%AE%BF%E9%97%AE%E5%B7%A5%E8%9C%82git-api
    """

    def __init__(self, oauth_token):
        self.oauth_token = oauth_token

    def __call__(self, r):
        r.headers["OAUTH-TOKEN"] = self.oauth_token
        return r



#!/usr/bin/env python
# encoding: utf-8
"""
@author: brill
@file: qbasenative.py
@time: 2019/7/22 10:35 AM
@desc:
"""
# !/usr/bin/env python3
# Created by xiazeng on 2019-05-22
from minium.native.wx_native.basenative import BaseNative


class QBaseNative(BaseNative):
    def __init__(self, json_conf):
        super(BaseNative, self).__init__()
        self.json_conf = json_conf

    def open_remote_debug(self, scheme_url):
        raise NotImplementedError()

    def release(self):
        raise NotImplementedError()

    def start_qq(self):
        """
        启动QQ
        :return:
        """
        raise NotImplementedError()

    def stop_qq(self):
        """
        启动QQ
        :return:
        """
        raise NotImplementedError()


class NativeError(RuntimeError):
    pass



#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Author:         lockerzhang
Filename:       connection_new.py
Create time:    2019/6/14 16:29
Description:

"""
import types
import typing
from typing_extensions import Self
import time
import websocket
from websocket import WebSocketConnectionClosedException
import json
from typing import TypeVar
from .minium_log import MonitorMetaClass
from ...framework.exception import *
from uuid import uuid4
import threading
import logging
from asyncio.coroutines import iscoroutine
from ...utils.utils import ProcessSafeEventLoop, WaitThread
from ...utils.emitter import ee
from ...utils.meta import PropertyMeta
from .callback import Callback

CLOSE_TIMEOUT = 5
MAX_WAIT_TIMEOUT = 30
MAX_RETRY = 3
g_thread = None
logger = logging.getLogger("minium")


class DevToolMessage(dict):
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise MiniNotAttributeError(name)

class BaseCommand(object, metaclass=PropertyMeta):
    _id = None
    """
    通信命令
    :method: 命令方法
    :params: 命令参数
    :desc: 命令超时时提醒信息
    """

    def __init__(
        self,
        method: str,
        params: dict = None,
        desc: str = None,
    ) -> None:
        self._id = None
        self.method = method
        self.params = params or {}
        self.desc = desc or f"call {method}"
        self._cancel = False

    """
    with BaseCommand(method, params) as cmd:
        cmd.id = "xxx"
        send(cmd)
    """
    def __enter__(self):
        return self
    
    def __exit__(self, ext_type, ext_value, ext_tb):
        self._cleanup(ext_type, ext_value, ext_tb)

    def __del__(self):
        self._cleanup()

    def _cleanup(self, *args):
        pass

    def dumps(self):
        return json.dumps(
            {"id": self.id, "method": self.method, "params": self.params},
            separators=(",", ":"),
        )
    
    @property
    def is_cancel(self):
        return self._cancel

    def cancel(self):
        self._cancel = True


class Command(BaseCommand):
    max_timeout = 30
    _id = None
    _has_response = False
    """
    通信命令
    :method: 命令方法
    :params: 命令参数
    :timeout: 命令返回时间，仅对同步指令有效
    :desc: 命令超时时提醒信息
    """

    def __init__(
        self,
        method: str,
        params: dict = None,
        max_timeout: int = None,
        desc: str = None,
    ) -> None:
        super().__init__(method, params, desc)
        self.max_timeout = max_timeout or Command.max_timeout
        self._has_response = False
        self.reason: Exception = None

    def on_connect_state_change(self, v, reason=MiniConnectionClosedError("connection break")):
        if not v:  # 下线
            self.reason = reason

    def on_id_change(self, value):
        """有设置id的情况, 代表指令是有"回复"的

        :param string v: id
        """
        self.has_response = True

    def on_has_response_change(self, value):
        if value:
            ee.on("connect_state_change", self.on_connect_state_change)
        else:
            ee.remove_listener("connect_state_change", self.on_connect_state_change)

    def _cleanup(self, *args):
        if self.has_response:
            self.has_response = False

    @classmethod
    def set_timeout(cls, timeout):
        cls.max_timeout = timeout


class AsyncCommand(BaseCommand):
    """异步命令"""
    def __init__(self, method: str, params: dict = None, desc: str = None) -> None:
        super().__init__(method, params, desc)
        self.result: DevToolMessage or Exception = None  # 命令返回结果

C = TypeVar('C', Command, AsyncCommand)

def json2obj(data):
    try:
        return json.loads(data, object_hook=DevToolMessage)
    except (TypeError, json.JSONDecodeError):
        return None


class Connection(object, metaclass=MonitorMetaClass):
    INSTANCES: typing.Dict[str, Self] = {}
    @classmethod
    def create(cls, uri, *args, **kwargs):
        """根据instance.id控制生成单例

        :param str uri: uri
        :return Connection: 实例
        """
        instance = cls(uri, *args, **kwargs)
        instance_id = instance.id
        if instance_id in cls.INSTANCES:
            old_instance = cls.INSTANCES.pop(instance_id)
            old_instance.destroy()
            old_instance = None
        cls.INSTANCES[instance_id] = instance
        return instance

    @classmethod
    def delete(cls, instance_id: str):
        if instance_id in cls.INSTANCES:
            del cls.INSTANCES[instance_id]

    def __init__(self, uri, timeout=MAX_WAIT_TIMEOUT):
        super().__init__()
        self._id = str(id(self))
        self.logger = logger.getChild(f"Conn{self._id[-4:]}")
        self.observers = {}
        self.uri = uri
        self.timeout = timeout
        Command.set_timeout(self.timeout)
        self.__is_connected = False
        self.__close_reason: Exception = None  # None for connecting
        self._connect_time = time.time()
        self._last_reconnect_fail_time = time.time() - MAX_WAIT_TIMEOUT  # 最后一次重试失败时间. 用于快速返回失败结果
        self._last_reconnect_fail_reason = None
        self._is_close_on_my_own = False  # 链接是自己主动断开的, 主要标记是否进行链接重连
        self._is_close_by_cmd = False  # 通过命令关闭的, 标记是否进行链接重连 和 小程序重新拉起
        self._is_close_app_by_cmd = False  # 通过app.exit命令关闭小程序，标记是否进行小程序重新拉起
        self._is_reconnecting = threading.RLock()
        self._msg_lock = threading.Condition()
        self._ws_event_queue = dict()
        self._req_id_counter = int(time.time() * 1000) % 10000000000
        self._sync_wait_map = {}
        self._async_msg_map: dict[str, AsyncCommand] = {}
        self._method_wait_map = {}
        self._thread = None
        # event loop 用来处理on message回调中的异步函数
        self._event_loop = ProcessSafeEventLoop()
        self._connect()

    @property
    def id(self):
        """一个链接的唯一标识

        :return str: unique id
        """
        return self.uri

    @property
    def _is_connected(self):
        return self.__is_connected

    @_is_connected.setter
    def _is_connected(self, value):
        last_value = self.__is_connected
        self.__is_connected = value
        if last_value != value:
            ee.emit("connect_state_change", value, self.__close_reason)
            self.connect_state_change(value)
            if not value:  # 断连
                self._set_all_async_command_fail(self.__close_reason)

    @property
    def is_close_by_myself(self):
        """
        是不是自己关闭的连接，如果是，不需要进行自动重连处理
        """
        if (
            self._is_close_app_by_cmd
            or self._is_close_by_cmd
            or self._is_close_on_my_own
        ):
            return True
        return False
    
    @property
    def is_close_conn_by_myself(self):
        return self._is_close_by_cmd or self._is_close_on_my_own
    
    @property
    def is_close_app_by_myself(self):
        return self._is_close_app_by_cmd or self._is_close_by_cmd
    
    @property
    def is_reconnecting(self):
        if self.is_close_by_myself:
            return False
        if self._is_reconnecting.acquire(False):
            self._is_reconnecting.release()
            return False
        return True

    def __del__(self):
        # 清理异步命令存储
        for k in list(self._async_msg_map.keys()):
            self._async_msg_map.pop(k)
        self._is_close_on_my_own = True
        self._is_close_by_cmd = True
        self._is_close_app_by_cmd = True

    def register(self, method: str, callback):
        if method not in self.observers:
            self.observers[method] = []
        self.observers[method].append(callback)

    def remove(self, method: str, callback=None):
        if method in self.observers.keys():
            if callback is None:  # remove all callback
                del self.observers[method]
            elif callback in self.observers[method]:
                self.observers[method].remove(callback)
        else:
            self.logger.debug("remove key which is not in observers")

    def remove_all_observers(self):
        try:
            obs_list = [x for x in self.observers.keys()]
            for obs in obs_list:
                del self.observers[obs]
        except Exception as e:
            raise KeyError(e)

    def create_async_callback_task(self, callback: types.FunctionType, *args):
        # self.logger.warn("create_async_callback_task: %s" % callback.__name__)
        async def _callback(*_args):
            # self.logger.warn("@async call %s" % callback.__name__)
            ret = callback(*_args)
            if iscoroutine(ret):
                return await ret
            return ret

        self._event_loop.run_coroutine(_callback(*args))

    def notify(self, method: str, message):
        if method == "App.bindingCalled" and message["name"] in self.observers:
            for callback in self.observers[message["name"]]:
                if callable(callback):
                    # callback(message)
                    self.create_async_callback_task(callback, message)
                else:
                    raise MiniNoncallableError(f"{str(callback)}(message={message})")
            return
        elif method in self.observers:
            for callback in self.observers[method]:
                if callable(callback):
                    # callback(message)
                    self.create_async_callback_task(callback, message)
                else:
                    raise MiniNoncallableError(f"{str(callback)}(message={message})")
        else:
            return

    def _connect(self, timeout=MAX_WAIT_TIMEOUT):
        error_callback = Callback()
        close_callback = Callback()
        ee.once("ws_close", close_callback.callback)
        ee.once("ws_error", error_callback.callback)
        self._client = websocket.WebSocketApp(
            self.uri,
            on_open=self._on_open,
            on_message=self._on_message,
            on_error=self._on_error,
            on_close=self._on_close,
        )
        self._thread = WaitThread(target=self._ws_run_forever, args=())
        self._thread.daemon = True
        self._thread.start()

        s = time.time()
        while (time.time() - s < timeout) and self._thread.is_alive():
            if self._is_connected:
                self.logger.info("connect to WebChatTools successfully")
                self._connect_time = time.time()
                break
        else:
            if (
                not self._thread.is_alive()
                and self._thread.get_result(block=False) is True
            ):
                if error_callback.wait_called(0):  # on_error早于run_forever结束
                    raise error_callback.result
                raise MiniConnectSvrError("connect websocket server exception")
            raise MiniTimeoutError(
                "connect to server timeout: %s, thread:%s"
                % (self.uri, self._thread.is_alive())
            )

    def _ws_run_forever(self):
        try:
            return self._client.run_forever()
        except Exception as e:
            self.logger.exception(e)
            self.__close_reason = MiniConnectionClosedError(str(e))
            self._is_connected = False
            return True
        finally:
            self.logger.info("websocket run forever shutdown")

    def _send(self, message):
        """发送消息

        :param str message: 消息
        :return bool: true 发送成功, false 发送失败
        """
        if self._client and self._client.sock:
            try:
                self._client.send(message)
                return True
            except WebSocketConnectionClosedException as e:
                if self.reconnect(time.time()):
                    self.logger.debug("RESEND > %s" % message)
                    self._client.send(message)
                    return True
                return False
        elif self._client and not self._client.sock:  # 断连了
            if self.reconnect(time.time()):
                self.logger.debug("RESEND > %s" % message)
                self._client.send(message)
                return True
            return False
        return False

    def _wait(self, cmd: Command):
        stime = time.time()
        self._msg_lock.acquire()
        self._msg_lock.wait(cmd.max_timeout)  # 如果是因为其他命令的返回触发了notify，需要重新等待
        self._msg_lock.release()
        etime = time.time()
        if etime - stime >= cmd.max_timeout:
            cmd.max_timeout = 0
        else:
            cmd.max_timeout = cmd.max_timeout - (etime - stime)  # 剩余要等待的时间

    def _notify(self):
        self._msg_lock.acquire()
        self._msg_lock.notify_all()  # 全部唤醒，让其自己决定是否需要重新wait
        self._msg_lock.release()

    def _safely_send(self, cmd: Command, sync=True):
        """发送前检查链接状态

        :param Command message: 命令
        :return None:
        :raise MiniConnectionClosedError:
        """
        message = cmd.dumps()
        self.logger.debug(f'{"SEND" if sync else "ASYNC_SEND"} > {message}')
        if not self._send(message) and not self._is_connected:
            raise MiniConnectionClosedError("send message[%s] fail because connection is not established" % cmd.id)

    def _gen_command(self, method: str or C, params=None, max_timeout=None, sync=True) -> C:
        if sync:
            if isinstance(method, Command):
                cmd = method
            elif isinstance(method, AsyncCommand):
                cmd = Command(method.method, method.params, desc=method.desc)
            else:
                cmd = Command(method, params, max_timeout)
        else:
            if isinstance(method, AsyncCommand):
                cmd = method
            elif isinstance(method, Command):
                cmd = AsyncCommand(method.method, method.params, desc=method.desc)
                del method  # 删除监听函数
            else:
                cmd = AsyncCommand(method, params)
        if not cmd.id:
            cmd.id = str(uuid4())
        return cmd

    def send(self, method: str or Command, params=None, max_timeout=None):
        # 同步发送消息，函数会阻塞
        cmd: Command = self._gen_command(method, params, max_timeout)
        with cmd:
            self._sync_wait_map[cmd.id] = None  # 该ID未有返回message
            self._safely_send(cmd)
            return self._receive_response(cmd)

    def send_async(self, method: str or AsyncCommand, params=None, ignore_response=False) -> str:
        cmd: AsyncCommand = self._gen_command(method, params, sync=False)
        if not ignore_response:
            self._async_msg_map[cmd.id] = cmd
        try:
            self._safely_send(cmd, sync=False)
        except MiniConnectError:
            if not ignore_response:
                self._async_msg_map.pop(cmd.id)
            del cmd
            raise
        return cmd.id

    def _receive_response(self, cmd: Command):
        # 等待接收到message的通知
        while cmd.max_timeout > 0:
            self._wait(cmd)

            if (
                cmd.id in self._sync_wait_map
                and self._sync_wait_map[cmd.id] is not None
            ):  # 获取到了数据
                response = self._sync_wait_map.pop(cmd.id)
                if isinstance(response, MiniAppError):
                    raise response
                return response
            elif cmd.method.startswith("Tool.") and self.is_close_conn_by_myself:
                if cmd.reason:
                    raise MiniConnectionClosedError("close by myself") from cmd.reason
                raise MiniConnectionClosedError("close by myself")
            elif self.is_close_app_by_myself:
                raise MiniClientOfflineError("client close by myself")
            elif cmd.max_timeout > 0:  # 线程是被其他消息唤醒，重新等待
                self.logger.debug("rewait for %s" % cmd.id)
                continue
            else:
                if cmd.id in self._sync_wait_map:
                    del self._sync_wait_map[cmd.id]
                if cmd.reason and isinstance(cmd.reason, MiniConnectionClosedError):
                    raise MiniTimeoutCauseByConnectionBreakError(
                        f"[{cmd.id}][{cmd.desc}] command timeout cause by {cmd.reason}"
                    )
                raise MiniTimeoutError(
                    f"[{cmd.id}][{cmd.desc}] receive from remote timeout"
                )

    def before_reconnect(self):
        time.sleep(0.1)

    def after_reconnect(self):
        pass

    def connect_state_change(self, value):
        if value is False and not self.is_close_by_myself:  # 原本已经链接上了, 并且不是自己关闭的链接
            threading.Thread(
                target=self.reconnect,
                args=(time.time(),),
                daemon=True,
                name="reconnect cause close",
            ).start()

    def _handle_connection_refused(self, retry):
        """connection refused对于不同环境需要有不同处理

        :param int retry: 当前重试次数
        """
        if retry >= MAX_RETRY:
            ee.emit("ide_closed", time.time())  # 多次refused, 并出现在reconnect中, 很可能是ide被close了
            raise MiniConnectSvrError("ConnectionRefused")
        time.sleep(5)

    def reconnect(self, optime=None, *args):
        RETRY_TIME = MAX_RETRY * MAX_WAIT_TIMEOUT + time.time()  # 最大重连尝试时间
        if self._is_close_on_my_own or self._is_close_by_cmd:
            return False
        if optime and optime < self._connect_time:
            return True
        if optime and optime < (self._last_reconnect_fail_time + MAX_WAIT_TIMEOUT):
            self.logger.warning("reconnect fast return False")
            # 加一个重连的冷却时间
            return False      
        # 等待其他线程重连
        if not self._is_reconnecting.acquire(False):
            t = threading.currentThread()
            self.logger.warning(
                "thread[%s:%s] reconnecting, wait for reconnection"
                % (t.getName(), t.ident)
            )
            if not self._is_reconnecting.acquire(
                timeout=MAX_RETRY * MAX_WAIT_TIMEOUT + 10
            ):  # 理论上应该能获取到锁的
                return False
            self.logger.warning(
                "another thread reconnect finish connected: " + str(self._is_connected)
            )
            self._is_reconnecting.release()
            if not self._is_connected:
                raise MiniReConnectSvrError("reconnect fail")
            return True
        if self._client:
            self._client.close()  # 先close一下
        retry = 0
        try:
            self.logger.warning("connection reconnect")
            self._last_reconnect_fail_reason = None
            self.before_reconnect()
            while time.time() < RETRY_TIME and not self.is_close_by_myself:
                try:
                    self._is_connected = False
                    self._connect(MAX_WAIT_TIMEOUT)
                    if self._is_connected:
                        return True
                    self.logger.warning("connection reconnect fail")
                    self._last_reconnect_fail_time = time.time()
                    self._last_reconnect_fail_reason = MiniTimeoutError("connect timeout")
                    return False
                except ConnectionRefusedError as cre:
                    self._last_reconnect_fail_reason = cre
                    retry += 1
                    self.logger.warning("connect refused, please check svr")
                    if time.time() < RETRY_TIME:
                        self._handle_connection_refused(retry)
                    continue
                except MiniTimeoutError as te:
                    self._last_reconnect_fail_reason = te
                    retry += 1
                    if retry >= MAX_RETRY:
                        break
                    self.logger.error("wait onopen timeout")
                    continue
                except MiniConnectSvrError as mcse:
                    self._last_reconnect_fail_reason = mcse
                    retry += 1
                    self.logger.error(mcse)
                    if retry >= MAX_RETRY:
                        break
                    if time.time() < RETRY_TIME:
                        time.sleep(2)
                    continue
                except Exception as e:
                    self._last_reconnect_fail_reason = e
                    retry += 1
                    self.logger.exception("reconnect error: %s", str(e))
                    time.sleep(10)
                    continue
        except Exception:
            self.logger.warning("connection reconnect fail")
            self._last_reconnect_fail_time = time.time()
            return False
        finally:
            self._is_reconnecting.release()
            self.after_reconnect()
        self.logger.warning("connection reconnect fail")
        self._last_reconnect_fail_time = time.time()
        return False

    def wait_reconnect(self, timeout=None) -> None or Exception:
        wait_time = MAX_RETRY * MAX_WAIT_TIMEOUT + 10 if timeout is None else timeout
        if not self._is_reconnecting.acquire(timeout=wait_time):
            return TimeoutError(f"connection did not reconnect within {wait_time} seconds")
        self._is_reconnecting.release()
        if not self._is_connected:
            return self._last_reconnect_fail_reason or MiniReConnectSvrError("reconnect fail")

    def _on_close(self, *args):
        if args and isinstance(args[0], websocket.WebSocketApp):
            args = args[1:]
        ee.emit("ws_close", args)
        self.__close_reason = MiniConnectionClosedError(
            "connection close: code[%s] reason[%s]" % args
            if len(args) == 2
            else "connection close"
        )
        self._is_connected = False

    def _on_error(self, error, *args):
        if args:
            # 会传 ws 实例的情况
            error = args[0]
        ee.emit("ws_error", error)
        if "Connection is already closed" in str(error):
            self.logger.warning(error)
            return
        self.logger.error(error)
        self.__close_reason = MiniConnectionClosedError("connection error: " + str(error))
        self._is_connected = False

    def _on_open(self, *args):
        self.__close_reason = None
        self._is_connected = True

    def _on_message(self, message, *args):
        """
        1. _on_message 调用都会在同一个线程
        2. 线程的阻塞不会导致接收消息失败, 但会影响处理消息速度(即消息不会丢, 但会慢)
        3. 有些ws的库会传ws实例！至今不懂为什么有些会有些不会，先兼容一下
        """
        if args:
            # 会传 ws 实例的情况
            message = args[0]
        if len(message) > 512:
            self.logger.debug("RECV < %s..." % message[:509])
        else:
            self.logger.debug("RECV < %s" % message)
        ret_json = json2obj(message)
        if isinstance(ret_json, dict):
            if "id" in ret_json:  # response
                req_id = ret_json["id"]
                if "error" in ret_json and "message" in ret_json["error"]:
                    err_msg = ret_json["error"]["message"]
                    if err_msg:  # 生成错误实例
                        self.logger.error(f"[{req_id}]: {err_msg}")
                        if err_msg.find("unimplemented") > 0:
                            self.logger.warn(
                                f"{err_msg}, It may be caused by the low sdk version"
                            )
                            ret_json = MiniLowVersionSdkError(err_msg)
                        elif err_msg == "Remote debug connection lost":
                            if self._is_close_app_by_cmd:
                                # 主动关闭小程序可能会导致丢失链接
                                ret_json = json2obj('{"id":"%s","result":{}}' % req_id)
                            raise RemoteDebugConnectionLost(err_msg)
                        else:
                            ret_json = MiniAppError(err_msg)
                if req_id in self._sync_wait_map:
                    self._sync_wait_map[req_id] = ret_json
                    self._notify()
                else:
                    self._handle_async_msg(req_id, ret_json)
            else:
                if "method" in ret_json and ret_json["method"] in self._method_wait_map:
                    self._method_wait_map[ret_json["method"]] = ret_json
                    self._notify()

                if "method" in ret_json and "params" in ret_json:
                    # self._push_event(ret_json["method"], ret_json["params"])
                    self.notify(ret_json["method"], ret_json["params"])

    def _handle_async_msg(self, req_id, ret_json):
        self.logger.info("received async msg: %s%s", req_id, "" if req_id in self._async_msg_map else ", maybe command ignore response")
        if ret_json is None:
            self.logger.warning("async msg[%s] response is None" % req_id)
        if ee.emit(req_id, ret_json):  # 有监听回调
            if req_id in self._async_msg_map:
                self._async_msg_map.pop(req_id)
        elif req_id in self._async_msg_map:  # 是这个实例发出的指令
            self._async_msg_map[req_id].result = ret_json

    def _set_all_async_command_fail(self, error: Exception):
        for cid in list(self._async_msg_map.keys()):
            self._handle_async_msg(cid, error)

    def _push_event(self, method, params):
        if method in self._ws_event_queue:
            self._ws_event_queue[method].append(params)
        else:
            self._ws_event_queue[method] = [params]

    def destroy(self):
        if self._is_close_on_my_own:
            self.logger.debug("already destroy")
            return
        self.logger.error("断开连接")
        self._is_close_on_my_own = True
        self._client.close()
        self._notify()
        self._set_all_async_command_fail(MiniConnectionClosedError("connection destroy"))
        if self._thread:
            self._thread.join(CLOSE_TIMEOUT)
        if self._event_loop.is_running():
            self._event_loop.stop_loop()
        self.__class__.delete(self.id)

    def wait_for(self, method: str or Command, max_timeout=None):
        if isinstance(method, Command):
            cmd = method
        else:
            cmd = Command(method, max_timeout=max_timeout)
        if cmd.method in self._method_wait_map and self._method_wait_map.get(
            cmd.method, None
        ):  # 已经有间听到并没被删除（一般只有别的线程同样在等这个方法，但又还没响应才会命中，兜底逻辑）
            return True
        self._method_wait_map[cmd.method] = None
        try:
            while cmd.max_timeout > 0:
                self._wait(cmd)
                if cmd.method in self._method_wait_map and self._method_wait_map.get(
                    cmd.method, None
                ):
                    del self._method_wait_map[cmd.method]
                    return True
                if cmd.is_cancel:
                    return False
                if cmd.method.startswith("Tool.") and not self.is_close_conn_by_myself:
                    continue
                if self.is_close_by_myself:
                    return False
            self.logger.error("Can't wait until %s" % cmd.method)
            return False
        finally:
            del cmd

    def get_aysnc_msg_return(self, msg_id=None):
        if not msg_id:
            self.logger.warning(
                "Can't get msg without msg_id, you can get msg_id when calling send_async()"
            )
            return None
        if msg_id in self._async_msg_map and isinstance(self._async_msg_map[msg_id], AsyncCommand):
            response = self._async_msg_map[msg_id].result
            if response is not None:
                self._async_msg_map.pop(msg_id)
            if isinstance(response, (MiniAppError, MiniConnectError)):
                raise response
            return response
        return None



#!/usr/bin/env python3
# Created by xiazeng on 2019-05-06
import datetime
import json
import os.path
from posixpath import realpath
import shutil
import time
import inspect

import minium
import minium.miniprogram.base_driver.minium_log
import minium.native
from minium.native.exception import *
from .miniconfig import MiniConfig
from .assertbase import AssertBase, HookAssert
from .logcolor import *
from .exception import *
from ..utils.utils import retry, catch
from ..utils.injectjs import getInjectJsCode
from .miniprogram import MiniProgram
from ..miniprogram import Minium
from ..miniprogram.base_driver.app import App
from ..native import BaseNative
from typing import Tuple

# import matplotlib.pyplot as plt

logger = logging.getLogger("minium")

g_minium: Minium = None
g_native: BaseNative = None
g_log_message_list = []
g_network_message_dict = {}  # 记录请求消息
g_network_req_cache = {}  # 记录请求体消息(降低重复请求消息量)
g_network_resp_cache = {}  # 记录请求返回消息(降低重复请求消息量)
FRAMEWORK_RETRY = 1  # case失败, 框架针对错误进行重试的次数


def full_reset():
    """
    在整个测试流程结束之后调用
    1. miniProgram 部分同 reset_minium()
    2. native 部分
        1.1 关闭微信
        1.2 释放原生 driver
    :return:
    """
    global g_minium, g_native
    if g_minium:
        reset_minium()
    if g_native:
        try:
            g_native.stop_wechat()
        except:
            pass
        try:
            g_native.release()
        except:
            pass
        g_native = None


def reset_minium():
    """
    1. miniProgram 部分 release
        1.1 非 ide 运行退出小程序
    2. 释放所有观察者的事件监听
    3. 销毁与 ide 的连接
    :return:
    """
    global g_minium
    if g_minium:
        g_minium.release()
    g_minium = None


def get_native(cfg: MiniConfig) -> BaseNative:
    """
    native 部分完全由配置中的 platform 参数控制
    为了确保不会过多地重启微信，已存在 g_native 则不创建,
    teardown 的时候也不会 release
    :param cfg: 配置
    :return:
    """
    global g_native
    if g_native is None:
        g_native = minium.native.get_native_driver(cfg.platform, cfg)
        g_native.start_wechat()
    return g_native


def get_minium(cfg: MiniConfig) -> Minium:
    """
    miniProgram 部分每次 classTearDown 是否 release 由配置 close_ide 控制
    初始化有如下工作：
        1. 处理配置
        2. 如果配置了 project_path 则启动 ide
        3. 连接 ide
        4. 根据配置 platform 判断是否需要远程调试
        5. 根据配置 enable_app_log 判断是否需要监听小程序 log 输出
        6. 根据配置 enable_network_panel 判断是否需要监听小程序 request, downloadFile, uploadFile 输出
    :param cfg: 配置
    :return:
    """
    global g_minium, g_native
    if g_minium is not None and (
        cfg.close_ide or cfg.full_reset
    ):  # 如果是配置了close_ide的，先release minium
        reset_minium()
    if g_minium is None:
        g_minium = minium.miniprogram.get_minium_driver(conf=cfg, native=g_native)

        if not cfg.use_push and cfg.platform != "ide":
            g_native.connect_weapp(g_minium.qr_code)
            ret = g_minium.connection.wait_for(method="App.initialized")
            if ret is False:
                raise MiniAppError("Launch MiniProgram Fail")

        if cfg.enable_app_log:
            g_minium.connection.register("App.logAdded", mini_log_added)
            g_minium.app.enable_log()

        if cfg.enable_network_panel:
            # 没有uuid库，随便注入一个随机ID库
            g_minium.app._evaluate_js("uuid")
            g_minium.app.expose_function("mini_request_callback", request_callback)
            g_minium.app.expose_function("mini_send_request", send_request)
            g_minium.app._evaluate_js("networkPannel")

        if cfg.platform == "ide" and g_native:
            g_native.mini = g_minium  # ide native接口通过minium实现

    return g_minium


def init_miniprogram(cfg: MiniConfig) -> Tuple[BaseNative, Minium, MiniProgram]:
    """
    初始化小程序
    1. 初始化native
    2. 启动微信
    3. 拉起小程序
    4. 获取必要的小程序信息
    """
    logger.info("start init miniprogram")
    native = get_native(cfg)
    mini = get_minium(cfg)
    if cfg.appid and MiniProgram.get_instance(cfg.appid):
        logger.info("miniprogram exists")
        return native, mini, MiniProgram.get_instance(cfg.appid)
    account_info = mini.get_app_config("accountInfo").accountInfo
    system_info = mini.get_system_info()
    appid = ""
    appname = ""
    if account_info:
        appid = account_info.get("appId", "")
        appname = account_info.get("appName", "") or account_info.get(
            "nickname", ""
        )  # 发现appname改成了nickname
    logger.info("end init miniprogram")
    return (
        native,
        mini,
        MiniProgram(appid or cfg.appid, appname=appname, system_info=system_info),
    )  # 返回一个默认的


def mini_log_added(message):
    """
    小程序 log 监听回调函数
    将小程序的 log 格式化然后保存起来
    :param message:
    :return:
    """
    dt = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    message["dt"] = dt
    g_log_message_list.append(message)


def send_request(message):
    [msg_id, obj, ms, hash_id] = message["args"]
    if hash_id and obj:  # 传了原始request obj, 记录起来
        g_network_req_cache[hash_id] = {"obj": obj, "timestamp": time.time()}
    elif hash_id and hash_id in g_network_req_cache:
        obj = g_network_req_cache[hash_id]["obj"]
        g_network_req_cache[hash_id]["timestamp"] = time.time()
    if msg_id not in g_network_message_dict:
        g_network_message_dict[msg_id] = {"timestamp": time.time() * 1000}
    g_network_message_dict[msg_id]["start_timestamp"] = ms
    g_network_message_dict[msg_id]["request"] = json.loads(obj)


def request_callback(message):
    [msg_id, res, ms, hash_id, mocked] = message["args"]
    if hash_id and res:  # 传了原始response res, 记录起来
        g_network_resp_cache[hash_id] = {"res": res, "timestamp": time.time()}
    elif hash_id and hash_id in g_network_resp_cache:
        res = g_network_resp_cache[hash_id]["res"]
        g_network_resp_cache[hash_id]["timestamp"] = time.time()
    if msg_id not in g_network_message_dict:
        g_network_message_dict[msg_id] = {"timestamp": time.time() * 1000}
    g_network_message_dict[msg_id]["end_timestamp"] = ms
    g_network_message_dict[msg_id]["response"] = json.loads(res)
    g_network_message_dict[msg_id]["mocked"] = bool(mocked)

class MetaMiniTest(HookAssert):
    def __new__(cls, name, bases, attrs):
        return super(MetaMiniTest, cls).__new__(cls, name, bases, attrs)

    @property
    def app(cls):
        if not inspect.isclass(cls):
            cls = type(cls)
        return cls._app or (cls.mini and cls.mini.app)

    @app.setter
    def app(cls, value):
        if not inspect.isclass(cls):
            cls = type(cls)
        cls._app = value


class MiniTest(AssertBase, metaclass=MetaMiniTest):
    mini: Minium = None
    native: BaseNative = None
    appId = ""
    appName = ""
    logger = logger

    _app: App = None

    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        # 这个case是否需要重试, 目前重试原则为
        # 1. case运行过程中遇到断连的情况
        self.__will_retry = None  

    @property
    def app(self) -> minium.App:
        return self._app or (self.mini and self.mini.app)

    @app.setter
    def app(self, value):
        self._app = value

    @classmethod
    def _miniClassSetUp(cls):
        logger.debug("=====================")
        logger.debug("Testing class：%s" % cls.__name__)
        logger.debug("=====================")
        super(MiniTest, cls)._miniClassSetUp()
        if not cls.CONFIG.report_usage:
            minium.miniprogram.base_driver.minium_log.existFlag = 1

        native, mini, miniprogram = init_miniprogram(cls.CONFIG)
        cls.native = native
        cls.mini = mini
        if cls.CONFIG.need_perf:  # start get perf thread
            cls.native.start_get_perf(timeinterval=0.8)
        cls.DEVICE_INFO["system_info"] = miniprogram.system_info

    @classmethod
    def tearDownClass(cls):
        """
        1. 存在 g_minium
        2. 配置 close_ide=True
        3. ide 下不释放
        :return:
        """
        if cls.CONFIG.full_reset:
            logger.info("full reset")
            if cls.CONFIG.audits and cls.mini and hasattr(cls.mini, "stop_audits"):
                cls.mini.stop_audits(
                    os.path.join(cls.CONFIG.outputs, "audits_%s.html" % cls.__name__)
                )
            full_reset()
            cls.mini = None
            cls.native = None
            return
        if g_minium and cls.CONFIG.close_ide and cls.CONFIG.platform == "ide":
            logger.info("close ide and reset minium")
            if cls.CONFIG.audits and cls.mini and hasattr(cls.mini, "stop_audits"):
                cls.mini.stop_audits(
                    os.path.join(cls.CONFIG.outputs, "audits_%s.html" % cls.__name__)
                )
            reset_minium()
            cls.mini = None

    def _check_debug_connection(self) -> ResetError:
        """检查调试通道

        :return None or Exception: None for ok
        """
        if self.native:
            if not self.native.check_connection():
                self.logger.warning("check native connection error, relaunch app")
                return ResetError.RELAUNCH_APP
        if self.mini: # 小程序异常退出和链接断连都会自动重连和重新拉起
            if self.mini.is_app_relaunching:
                self.logger.warning("app is relaunching, wait for it")
                e = self.mini.wait_app_relaunch()
                if e:
                    self.logger.warning(e)
                    return ResetError.RELAUNCH_APP
            if self.mini.connection and self.mini.connection.is_reconnecting:
                self.logger.warning("connection is reconnecting, wait for it")
                e = self.mini.connection.wait_reconnect()
                if e:
                    self.logger.warning(e)
                    return ResetError.RELAUNCH_APP
        return ResetError.OK
    
    def _update_miniprogram(self, native, mini):
        self.logger.debug(
            "start update miniprogram, class is: %s" % self.__class__.__name__
        )
        self.__class__.native = native
        self.__class__.mini = mini
        self.mini = mini
        self.native = native
        self.logger.debug(
            "finish update miniprogram, current native: %s, minium: %s"
            % (id(self.mini), id(self.native))
        )

    def init_miniprogram(self):
        """case中重新实例化小程序
        """
        try:
            native, mini, _ = init_miniprogram(self.__class__.CONFIG)
        except:
            self._update_miniprogram(g_native, g_minium)
            raise
        else:
            self._update_miniprogram(native, mini)

    def _miniSetUp(self):
        super(MiniTest, self)._miniSetUp()
        self._is_perf_setup = False
        self._is_audits_setup = False
        logger.info("=========Current case: %s=========" % self._testMethodName)
        if self.__will_retry is not None:
            self.logger.warning(f"第{FRAMEWORK_RETRY-self.__will_retry}次重试")
        logger.info(
            "package info: %s, case info: %s.%s"
            % (
                self.results.get("module", ""),
                self.__class__.__name__,
                self._testMethodName,
            )
        )
        if self.test_config.only_native:
            logger.info(f"Only native: {self.test_config.only_native}, setUp complete")
            return
        else:
            check_result = self._check_debug_connection()
            if check_result != ResetError.OK:  # 调试通道不通
                self.logger.warning("check debug connection fail")
                if check_result == ResetError.RELAUNCH_APP:
                    full_reset()
                    self._update_miniprogram(None, None)
                elif check_result == ResetError.RELAUNCH_MINIPROGRAM:
                    reset_minium()
                    self._update_miniprogram(self.native, None)
                self.init_miniprogram()

            if check_result == ResetError.OK and self.test_config.auto_relaunch:
                ret = self.native and self.native.back_to_miniprogram()
                if ret and ret != ResetError.OK:
                    if ret == ResetError.RELAUNCH_APP:
                        self.logger.warning(
                            "back_to_miniprogram error, reset app, post native: %s, minium: %s"
                            % (id(self.mini), id(self.native))
                        )
                        full_reset()
                        self._update_miniprogram(None, None)
                        self.init_miniprogram()
                    elif ret == ResetError.RELAUNCH_MINIPROGRAM:
                        self.logger.warning(
                            "back_to_miniprogram error, reset miniprogram, post minium: %s"
                            % id(self.native)
                        )
                        reset_minium()
                        self._update_miniprogram(self.native, None)
                        self.init_miniprogram()
                self.app.go_home()
        # 体验评分是个整体的评价，每个case太短了，没意义
        # if self.test_config.audits:
        #     self._is_audits_setup = self._setup_audits()
        self._framework_capture("setup")
        self._is_perf_setup = self._setup_perf()
        # update start_timestamp
        self.results["start_timestamp"] = time.time()
        logger.info("=========case: %s start=========" % self._testMethodName)

    def _setup_perf(self):
        self._get_perf_flag = False
        self.perf_data = ""
        if self.test_config.need_perf:
            # logger.debug("_setup_perf")
            self._get_perf_flag = bool(self.native.start_get_perf(timeinterval=0.8))
        return True

    @catch
    def _teardown_perf(self):
        """
        落地性能相关数据
        """
        if not self._is_perf_setup:
            return
        perf_init = {
            "startup": self.native.get_start_up(),  # 启动时间
            "avg_cpu": 0,  # 平均CPU
            "max_cpu": 0,  # 最大CPU
            "cpu_data_list": [],  # cpu 数据，可以每3s 打一次
            "avg_mem": 0,  # 平均内存
            "max_mem": 0,  # 最大内存
            "mem_data_list": [],  # 内存数据，可以每3s 打一次
            "avg_fps": 0,  # 平均FPS, 小游戏才有
            "min_fps_rt": 0,  # 最小FPS, 小游戏才有
            "fps_data_list": [],  # FPS数据，可以每3s 打一次
            "fps_time_series_list": [],
            "cpu_time_series_list": [],
            "mem_time_series_list": [],
        }
        if self._get_perf_flag:
            self._get_perf_flag = False
            perf_str = self.native.get_perf_data(self.setup_time)
            if self.native.outputs_screen and os.listdir(self.native.outputs_screen):
                page_path = self.page.path
                for file_name in os.listdir(self.native.outputs_screen):
                    file_name_all = os.path.join(self.native.outputs_screen, file_name)
                    # logger.error(file_name_all)
                    shutil.copy(file_name_all, self.screen_dir)
                    path = os.path.join(self.screen_dir, file_name)
                    # logger.error(path)
                    file_name_noext = os.path.splitext(file_name)[0]
                    self.screen_info.append(
                        {
                            "name": file_name,
                            "url": page_path,
                            "path": self.get_relative_path(path),
                            "ts": int(file_name_noext),
                            "datetime": datetime.datetime.fromtimestamp(
                                int(file_name_noext)
                            ).strftime("%Y-%m-%d %H:%M:%S"),
                        }
                    )
            if not perf_str:
                self.perf_data = json.dumps(perf_init)
            else:
                try:
                    perf_arr = json.loads(perf_str)
                    perf_re = perf_init
                    timestamp_arr = []
                    cpu_arr = []
                    mem_arr = []
                    fps_arr = []
                    fps_time_series_list = []
                    cpu_time_series_list = []
                    mem_time_series_list = []
                    for item in perf_arr:
                        _timestamp = int(item["timestamp"])
                        timestamp_arr.append(_timestamp)
                        if item.get("cpu", None) is not None:
                            cpu_arr.append(int(item["cpu"]))
                            cpu_time_series_list.append(_timestamp)
                        if item.get("fps", None) is not None:
                            fps_arr.append(int(item["fps"]))
                            fps_time_series_list.append(_timestamp)
                        if item.get("mem", None) is not None:
                            mem_arr.append(float(item["mem"]))
                            mem_time_series_list.append(_timestamp)
                    perf_re["fps_time_series_list"] = fps_time_series_list
                    perf_re["cpu_time_series_list"] = cpu_time_series_list
                    perf_re["mem_time_series_list"] = mem_time_series_list
                    perf_re["cpu_data_list"] = cpu_arr
                    perf_re["mem_data_list"] = mem_arr
                    perf_re["fps_data_list"] = fps_arr
                    perf_re["avg_cpu"] = sum(cpu_arr) / (len(cpu_arr) or 1)
                    perf_re["avg_mem"] = sum(mem_arr) / (len(mem_arr) or 1)
                    perf_re["avg_fps"] = sum(fps_arr) / (len(fps_arr) or 1)
                    perf_re["max_cpu"] = max(cpu_arr)
                    perf_re["max_mem"] = max(mem_arr)
                    perf_re["min_fps_rt"] = min(fps_arr)
                    self.perf_data = json.dumps(perf_re)
                except:
                    self.perf_data = json.dumps(perf_init)

        else:
            self.perf_data = json.dumps(perf_init)
        self.results["perf_data"] = self.perf_data

    # 不应在case setup做
    def _setup_audits(self):
        self.audit_html = ""
        self.audit_json = ""
        self.start_audit = False
        if self.test_config.platform == "ide":
            logger.info(f"start run audits score")
            self.start_audit = True
            # 开始执行体验评分用例
        else:
            logger.warn(f"Only ide can run audits score")
        return True

    # 不应在case teardown做
    def _teardown_audits(self):
        """
        落地测评数据
        """
        if not self._is_audits_setup:
            return
        if self.start_audit:
            self.stop_audits()
        self.results["audit_html"] = self.audit_html
        self.results["audit_json"] = self.audit_json

    @catch
    def _teardown_app_log(self):
        """
        落地小程序相关的log
        """
        global g_log_message_list, g_network_message_dict
        # 落地小程序log
        weapp_path = "weapp.log"
        weapp_filename = self.wrap_filename(weapp_path)
        log_messages = g_log_message_list
        g_log_message_list = []
        with open(weapp_filename, "w", encoding="UTF-8") as f:
            for log_message in log_messages:
                f.write(json.dumps(log_message, ensure_ascii=False) + "\n")
        self.results["weapp_log_path"] = weapp_path

        # 落地网络请求log
        request_path = "request.log"
        request_filename = self.wrap_filename(request_path)
        network_message = g_network_message_dict
        g_network_message_dict = {}
        network_message_list = [network_message[msg_id] for msg_id in network_message]
        network_message_list.sort(
            key=lambda x: x.get("start_timestamp", None) or x["timestamp"]
        )
        with open(request_filename, "w", encoding="UTF-8") as f:
            # msg => {timestamp, start_timestamp, request: dict or None, end_timestamp, response: dict or None}
            for msg in network_message_list:
                # logger.debug("\nrequest: {}\nresponse: {}".format(msg.get("request", "Error: Request None"),
                # msg.get("response", "Error: Response Empty")))
                if not msg.get("start_timestamp"):
                    msg["start_timestamp"] = msg["timestamp"]
                    msg["request"] = None
                if not msg.get("end_timestamp"):
                    msg["end_timestamp"] = msg["timestamp"]
                    msg["response"] = None
                f.write(json.dumps(msg, ensure_ascii=False) + "\n")
        self.results["request_log_path"] = request_path

    def _teardown_snapshot(self):
        try:
            if self.test_config.teardown_snapshot:
                self.results["page_wxml"] = self.page_wxml
                self.results["page_data"] = self.page.data
        except Exception as e:
            logger.exception(e)
            self.results["page_data"] = None
            self.results["page_wxml"] = ""

    @catch
    def _check_case_error(self, error: Exception):
        """检查失败的case的错误, 决定是否需要重试或者获取更多信息

        :param Exception error: case运行产生的错误
        """
        if not error:
            return
        if isinstance(error, (MiniConnectionClosedError, MiniReConnectSvrError)):
            # 因为断连而产生的指令超时, 可尝试重试{1}次
            if self.__will_retry is None:
                self.__will_retry = FRAMEWORK_RETRY
            return
        if isinstance(error, MiniElementNotFoundError) and not self.test_config.teardown_snapshot:
            # 找不到元素, 把wxml拿回来帮助用户排查. 如果配置了`teardown_snapshot`, 则已经获取过不需要再获取
            self.results["page_wxml"] = self.page_wxml
            return

    def _miniTearDown(self):
        logger.debug("=========Current case Down: %s=========" % self._testMethodName)
        self._framework_capture("teardown")
        sys_info = self._get_error_info()[0]
        if sys_info:  # case失败了
            self._check_case_error(sys_info[1])
        # 更新小程序专属的数据
        self._teardown_perf()
        self._teardown_app_log()
        self._teardown_snapshot()
        super(MiniTest, self)._miniTearDown()

    def _cleanup_before_retry(self):
        # 清理一下输出路径(等于删除上一个结果)
        if self.test_config.case_output and os.path.isdir(self.test_config.case_output):
            shutil.rmtree(self.test_config.case_output)

    def run(self, result=None):
        ret = super().run(result)
        if self.__will_retry is None:
            return ret
        while self.__will_retry > 0:
            self.__will_retry -= 1
            self._cleanup_before_retry()
            self.logger.warning(f"第{FRAMEWORK_RETRY-self.__will_retry}次重试")
            ret = super().run(result)
        return ret

    @property
    def page(self) -> minium.Page:
        return self.mini.app.current_page

    @property
    def page_wxml(self) -> str:
        if not self.mini:
            return ""
        if self.mini.sdk_version > "2.19.5":
            wxml = self.page.get_element("/*").outer_wxml
        else:
            wxml = self.page.get_element("page").inner_wxml
        if wxml:
            filename = datetime.datetime.now().strftime("%d%H%M%S.wxml")
            filepath = self.wrap_filename(filename)
            with open(filepath, "w", encoding="utf8") as fd:
                fd.write(wxml)
            return filename
        return wxml
    
    @catch
    @retry(2)
    def _framework_capture(self, name=""):
        """框架截图, 尝试两次, 不报错

        :param str name: 截图文件名, defaults to ""
        """
        if not self.test_config.framework_capture:
            return ""
        return self.capture(name)
    
    @catch
    @retry(2)
    def _error_capture(self, error: Exception):
        """报错截图, 尝试两次, 不报错

        :param Exception error: 具体报错, 截图文件名由报错类名定义
        """
        if not self.test_config.error_capture:
            return ""
        return self.capture(error.__class__.__name__)

    def capture(self, name=""):
        """
        :param name: 图片名称
        :return: str: image_path or ""
        """
        if name:
            filename = "%s.png" % name
        else:
            filename = "%s.png" % datetime.datetime.now().strftime("%H%M%S%f")
        path = os.path.join(self.screen_dir, filename)
        if os.path.isfile(path):
            # 文件已经存在，重命名
            tmp = os.path.splitext(filename)
            filename = (
                tmp[0] + "_%s" % datetime.datetime.now().strftime("%H%M%S%f") + tmp[1]
            )
            path = os.path.join(self.screen_dir, filename)
        logger.info("capture %s" % filename)
        self.native.screen_shot(path)
        if os.path.exists(path):
            page_path = (
                self.mini.app.route_return_page.path
                if self.mini and self.mini.app.route_return_page
                else ""
            )
            self.add_screen(name, path, page_path)
            return path
        else:
            logger.warning("%s not exists", path)
        return ""

    def tap_capture(self, name, page_id):
        """
        点击截图
        :param name: 截图文件名
        :param page_id: 点击元素所属的页面id
        """
        raise NotImplementedError("仅云测支持")
    
    def tap_mark(self, name: str, point: Tuple, page_id):
        """
        给截图标记点击的点
        :return: bool
        """
        raise NotImplementedError("仅云测支持")

    def stop_audits(self, format=None):
        """
        获取体验评分
        :return: 体验评分报告
        """
        format = ["html", "json"] if format is None else format
        ret = self.app._stop_audits()
        if len(format) == 0:
            raise Exception("未定义format")
        if not "result" in ret:
            raise Exception("stop_audits获取数据为空")
        if not "report" in ret["result"]:
            raise Exception("stop_audits获取html数据为空")
        if not "data" in ret["result"]:
            raise Exception("stop_audits获取json数据为空")

        for item in format:
            if item == "html":
                audits_html_file = "Audits.html"
                audits_htmlname = self.wrap_filename(audits_html_file)
                html_result = ret["result"]["report"]
                with open(audits_htmlname, "w", encoding="utf-8") as h_f:
                    h_f.write(html_result)
                h_f.close()
                self.audit_html = audits_html_file
                logger.info("success create Audits.html")
            elif item == "json":
                audits_json_file = "Audits.json"
                audits_jsonname = self.wrap_filename(audits_json_file)
                json_result = ret["result"]["data"]
                with open(audits_jsonname, "w", encoding="UTF-8") as j_f:
                    j_f.write(json_result)
                j_f.close()
                self.audit_json = audits_json_file
                logger.info("success create Audits.json")
            else:
                pass

    # 录制回放专用接口
    def mark_step_start(self, key, **kwargs):
        """录制回放步骤标记
        1. 打日志
        2. 额外信息加入`result.step_info`中

        :param str key: 步骤键名
        """
        self.logger.debug("[minitest replay] %s" % key)
        self.results.step_info.append(
            {"key": key, "timestamp": int(time.time()), **kwargs}
        )

    # 小程序定制化的校验
    def assertPageData(self, data, msg=None):
        """

        :param data:
        :param msg:
        :return:
        """
        pass

    def assertContainTexts(self, texts, msg=None):
        pass

    def assertTexts(self, texts, selector="", msg=None):
        for text in texts:
            elem = self.page.get_element(selector, inner_text=text)
            if elem is None:
                raise AssertionError("selector:%s, inner_text=%s not Found")

    def hook_assert(self, name, ret, reason=None):
        if self.test_config.assert_capture:
            filename = "{0}-{1}".format(name, "success" if ret else "failed")
            retry(2)(self.capture)(filename)
            wxml = ""
            if ret is False:
                wxml = self.page_wxml
            return {"img": filename, "wxml": wxml}



class MiniProgram:
    class __MiniProgram(object):
        def __init__(self, appid, **kwargs) -> None:
            self.appid = appid
            self.update(**kwargs)

        def update(self, **kwargs):
            for k in kwargs:
                setattr(self, k, kwargs[k])

    instances = {}
    def __init__(self, appid, **kwargs):
        appid = appid or ""
        if not MiniProgram.instances.get(appid):
            MiniProgram.instances[appid] = MiniProgram.__MiniProgram(appid, **kwargs)
        else:
            MiniProgram.instances[appid].update(**kwargs)
        self.instance = MiniProgram.instances[appid]
        
    def __getattr__(self, name):
        return getattr(self.instance, name, None)

    @classmethod
    def get_instance(cls, appid):
        return MiniProgram.instances.get(appid or "")
    



'''
Author: yopofeng yopofeng@tencent.com
Date: 2023-03-27 16:29:57
LastEditors: yopofeng yopofeng@tencent.com
LastEditTime: 2023-03-27 16:30:14
FilePath: /py-minium/minium/miniprogram/base_driver/version.py
Description: minium build version
'''
import os
import json

def build_version():
    config_path = os.path.join(os.path.dirname(__file__), "version.json")
    if not os.path.exists(config_path):
        return {}
    else:
        with open(config_path, "r", encoding="utf8") as f:
            version = json.load(f)
            return version


#!/usr/bin/env python3
# Created by xiazeng on 2019-06-03
import json
import os.path
import sys
import glob
import shutil
import time
import re


def formatTimeFromTimestamp(format, seconds):
    # 如果format有中文，会有问题，作转码处理
    s = str(format.encode("utf-8"))
    rtn = time.strftime(s, time.localtime(seconds))
    return eval(rtn).decode("utf-8")

def formatCostTimeFromTimestamp(format, seconds):
    # 如果format有中文，会有问题，作转码处理
    s = str(format.encode("utf-8"))
    rtn = time.strftime(s, time.gmtime(seconds))
    return eval(rtn).decode("utf-8")


def getTimestampFromFormattedTime(s, format="%Y-%m-%d %H:%M:%S"):
    time_struct = time.strptime(s, format)
    return time.mktime(time_struct)


ONE_DAY = 86400
ONE_HOUR = 3600
ONE_MIN = 60
VERSION = None  # 1, 2, 3


def formatCostTime(seconds):
    if seconds > ONE_DAY:
        days = int(seconds / ONE_DAY)
        return "%d天" % days + formatCostTimeFromTimestamp("%H小时%M分钟%S秒", seconds % ONE_DAY)
    elif seconds > ONE_HOUR:
        return formatCostTimeFromTimestamp("%H小时%M分钟%S秒", seconds)
    elif seconds > ONE_MIN:
        return formatCostTimeFromTimestamp("%M分钟%S秒", seconds)
    else:
        return "%.3f秒" % seconds


def getFileLastModificationTime(file_path, format="%Y%m%d %H:%M:%S"):
    t = os.path.getmtime(file_path)
    timeStruce = time.localtime(t)
    return t, time.strftime(format, timeStruce)


def get_result_json(outputs):
    """
    获取结果json文件, 并检测是否存在"未完成的用例"的log文件
    :params outputs: path
    :return filenames: list
    :return unfinished_filenames: list
    """
    global VERSION
    # 获取v3: run.log文件，v2: ${_testMethodName}.json，需要兼容
    if VERSION is None:  # 检测报告数据版本
        filenames = glob.glob(os.path.join(outputs, "*", "*", "run.log"))
        if not filenames:
            # 尝试查找 ${_testMethodName}.json
            filenames = glob.glob(os.path.join(outputs, "*", "*", "*.json"))
            if not filenames:
                return filenames, []
            for filename in filenames:
                relpath = os.path.relpath(filename, outputs)
                test_method_name = relpath.split(os.path.sep)[0]
                if (
                    os.path.basename(filename) == "%s.json" % test_method_name
                ):  # 符合v2版本格式
                    VERSION = 2  # 版本退化
                    return get_result_json(outputs)
            # v2, v3格式都不符合
            return [], []
        else:
            VERSION = 3
            return get_result_json(outputs)
    elif VERSION == 3:
        result_json_filenames = []
        unfinished_filenames = []
        filenames = glob.glob(os.path.join(outputs, "*", "*", "run.log"))
        for filename in filenames:
            result_json_filename = os.path.join(
                os.path.dirname(filename), "result.json"
            )
            if os.path.isfile(result_json_filename):
                result_json_filenames.append(result_json_filename)
                continue
            # 有run.log但没有result.json，尝试生成一个失败结果
            unfinished_filenames.append(filename)
        return result_json_filenames, unfinished_filenames
    else:
        return glob.glob(os.path.join(outputs, "*", "*", "*.json")), []


def get_case_datas(filenames):
    """
    获取case datas
    """
    case_datas = []
    for filename in filenames:
        relative_path = os.path.sep.join(filename.split(os.path.sep)[-3:-1])
        m = json.load(open(filename, "r", encoding="utf8"))
        m["relative"] = relative_path
        # 查找同目录下 perfdata.json 和 fpsdata.json
        perf_filename = os.path.join(os.path.dirname(filename), "perfdata.json")
        fps_filename = os.path.join(os.path.dirname(filename), "fpsdata.json")
        if os.path.isfile(perf_filename):
            perf = json.load(open(perf_filename, "r", encoding="utf8"))
            m["performance"] = perf
        if os.path.isfile(fps_filename):
            fps = json.load(open(fps_filename, "r", encoding="utf8"))
            m["fps"] = fps
        case_datas.append(m)
    case_datas.sort(key=lambda a: a["start_timestamp"])
    return case_datas


def get_unfinished_case_datas(unfinished_filenames):
    unfinished_case_datas = {}  # {package_name: {class_name: {case_name: {**info}}}}
    for filename in unfinished_filenames:
        reg_package_info = r"package info: (.+), case info: (\S+)\.(\S+)"
        with open(filename, "r", encoding="utf8") as fd:
            for line in iter(fd.readline, b""):
                if not line:
                    break
                m = re.search(reg_package_info, line)
                if m:
                    package_name, class_name, case_name = m.groups()
                    if package_name in unfinished_case_datas:
                        package_info = unfinished_case_datas[package_name]
                    else:
                        package_info = {}
                        unfinished_case_datas[package_name] = package_info
                    if class_name in package_info:
                        class_info = package_info[class_name]
                    else:
                        class_info = {}
                        package_info[class_name] = class_info
                    if case_name in class_info:
                        break
                    # try to get start time in log: %Y-%m-%d %H:%M:%S
                    mm = re.search(r"\d+-\d+-\d+ \d+:\d+:\d+", line)
                    if mm:
                        start_timestamp = getTimestampFromFormattedTime(
                            mm.group(0), "%Y-%m-%d %H:%M:%S"
                        )
                        run_time = formatTimeFromTimestamp(
                            "%Y%m%d %H:%M:%S", start_timestamp
                        )
                    else:
                        start_timestamp, run_time = getFileLastModificationTime(
                            filename
                        )
                    stop_timestamp = getFileLastModificationTime(filename)[0]
                    case_info = {
                        "log_filename": os.path.basename(filename),
                        "test_type": class_name,
                        "relative": os.path.sep.join(
                            filename.split(os.path.sep)[-3:-1]
                        ),
                        "module": package_name,
                        "run_time": run_time,
                        "start_timestamp": start_timestamp,
                        "stop_timestamp": stop_timestamp,
                    }
                    class_info[case_name] = case_info
                    # try to record other log
                    log_paths = {}
                    log_files = glob.glob(
                        os.path.join(os.path.dirname(filename), "*.log")
                    )
                    for log_file in log_files:
                        log_name = os.path.basename(log_file)
                        if log_name == case_info["log_filename"]:  # run.log, pass
                            continue
                        name = os.path.splitext(log_name)[0]
                        log_paths[f"{name}_log_path"] = log_name
                    case_info.update(log_paths)
                    # try to process screen info
                    screen_info = []
                    image_files = (
                        glob.glob(
                            os.path.join(os.path.dirname(filename), "images", "*")
                        )
                        if os.path.isdir(
                            os.path.join(os.path.dirname(filename), "images")
                        )
                        else []
                    )
                    for image_file in image_files:
                        image_name = os.path.basename(image_file)
                        ts, datetime = getFileLastModificationTime(
                            image_file, "%Y-%m-%d %H:%M:%S"
                        )
                        screen_info.append(
                            {
                                "name": os.path.splitext(image_name)[0],
                                "url": "",
                                "path": "images/%s" % image_name,
                                "ts": ts,
                                "datetime": datetime,
                            }
                        )
                        screen_info.sort(key=lambda x: x["ts"])
                        case_info["screen_info"] = screen_info
    return unfinished_case_datas


def get_case_summary(case_datas, unfinished_case_datas=None):
    success = 0
    failed = 0
    error = 0
    costs = 0
    pkg_info = {}
    detail = []
    start_time = None
    end_time = None
    for c in case_datas:
        c["costs"] = formatCostTime(
            c.get("stop_timestamp", c["start_timestamp"]) - c["start_timestamp"]
        )
        if c["success"]:
            success += 1
        elif c["is_failure"]:
            failed += 1
        elif c["is_error"]:
            error += 1
        start_time = (
            c["start_timestamp"]
            if start_time is None
            else min(c["start_timestamp"], start_time)
        )
        end_time = (
            c.get("stop_timestamp", c["start_timestamp"])
            if end_time is None
            else max(c.get("stop_timestamp", c["start_timestamp"]), end_time)
        )
        costs = end_time - start_time
        if c.get("package"):
            if c["package"] in pkg_info:
                pkg_info[c["package"]]["case_num"] += 1
                if c["success"]:
                    pkg_info[c["package"]]["success"] += 1
                elif c["is_failure"]:
                    pkg_info[c["package"]]["failed"] += 1
                elif c["is_error"]:
                    pkg_info[c["package"]]["error"] += 1
                pkg_info[c["package"]]["start_time"] = min(
                    c["start_timestamp"], pkg_info[c["package"]]["start_time"]
                )
                pkg_info[c["package"]]["end_time"] = max(
                    c.get("stop_timestamp", c["start_timestamp"]),
                    pkg_info[c["package"]]["end_time"],
                )
                pkg_info[c["package"]]["costs"] = (
                    pkg_info[c["package"]]["end_time"]
                    - pkg_info[c["package"]]["start_time"]
                )
            else:
                pkg_info[c["package"]] = {
                    "name": c["package"],
                    "case_num": 1,
                    "success": 1 if c["success"] else 0,
                    "failed": 1 if c["is_failure"] else 0,
                    "error": 1 if c["is_error"] else 0,
                    "start_time": c["start_timestamp"],
                    "end_time": c.get("stop_timestamp", c["start_timestamp"]),
                    "costs": c.get("stop_timestamp", c["start_timestamp"])
                    - c["start_timestamp"],
                }
    if unfinished_case_datas:
        for package_name in unfinished_case_datas:
            for class_name in unfinished_case_datas[package_name]:
                for case_name in unfinished_case_datas[package_name][class_name]:
                    item = unfinished_case_datas[package_name][class_name][case_name]
                    if item.get("start_timestamp"):
                        start_time = (
                            item["start_timestamp"]
                            if start_time is None
                            else min(item["start_timestamp"], start_time)
                        )
                    if item.get("stop_timestamp"):
                        end_time = (
                            item.get("stop_timestamp", item["start_timestamp"])
                            if end_time is None
                            else max(item.get("stop_timestamp", item["start_timestamp"]), end_time)
                        )

    for package in pkg_info:
        pkg_info[package]["start_time"] = formatTimeFromTimestamp(
            "%Y/%m/%d %H:%M:%S", pkg_info[package]["start_time"]
        )
        pkg_info[package]["end_time"] = formatTimeFromTimestamp(
            "%Y/%m/%d %H:%M:%S", pkg_info[package]["end_time"]
        )
        pkg_info[package]["costs"] = formatCostTime(pkg_info[package]["costs"])
        detail.append(pkg_info[package])
    detail.sort(key=lambda a: a["start_time"])
    return success, failed, error, costs, start_time, end_time, detail


def generate_meta(outputs):
    """
    v2: ${outputs}/${self._testMethodName}/${datetime.datetime.now().strftime("%Y%m%d%H%M%S")}/
        (${self._testMethodName}.json|${self._testMethodName}.log)
    v3: ${outputs}/${self._testMethodName}/${datetime.datetime.now().strftime("%Y%m%d%H%M%S")}/(result.json|run.log)

    如果case没有正确跑完, 会出现有run.log但没有result.json的情况, 尝试生成一个失败结果
    """
    filenames, unfinished_filenames = get_result_json(outputs)
    if not filenames and not unfinished_filenames:
        return False
    # case datas
    case_datas = get_case_datas(filenames)

    # unfinished cases
    unfinished_case_datas = get_unfinished_case_datas(unfinished_filenames)

    # summary
    success, failed, error, costs, start_time, end_time, detail = get_case_summary(
        case_datas,
        unfinished_case_datas
    )

    summary = {
        "case_num": len(case_datas),
        "success": success,
        "failed": failed,
        "error": error,
        "costs": formatCostTime(costs),
        "start_time": formatTimeFromTimestamp("%Y/%m/%d %H:%M:%S", start_time),
        "end_time": formatTimeFromTimestamp("%Y/%m/%d %H:%M:%S", end_time),
        "detail": detail,
    }
    meta_json = {
        "case_datas": case_datas,
        "summary": summary,
        "unfinished_case_datas": unfinished_case_datas,
    }
    return meta_json


def generate_meta_v2(outputs):
    tasks = glob.glob(os.path.join(outputs, "[0123456789]*"))
    task_datas = []
    for task in tasks:
        task_data = generate_meta(task)
        if not task_data:
            continue
        for case_data in task_data["case_datas"]:
            case_data["relative"] = os.path.join(
                os.path.basename(task), case_data["relative"]
            )
        if "unfinished_case_datas" in task_data:
            for k in task_data["unfinished_case_datas"]:
                for kk in task_data["unfinished_case_datas"][k]:
                    for kkk in task_data["unfinished_case_datas"][k][
                        kk
                    ]:  # kkk == case name
                        case_data = task_data["unfinished_case_datas"][k][kk][kkk]
                        case_data["relative"] = os.path.join(
                            os.path.basename(task), case_data["relative"]
                        )
        task_datas.append(task_data)
    task_datas.sort(key=lambda a: a["summary"]["start_time"], reverse=True)

    summary = {}
    meta_json = {"version": 2, "task_datas": task_datas, "summary": summary}
    return meta_json


generate_meta_v3 = generate_meta_v2


# meta_json = generate_meta("/Users/mmtest/code/minium/py-sample/outputs")
# json.dump(meta_json, open("meta.json", "w"), indent=4)


def imp_main(input_path, output_path=None):
    if output_path is None:
        output_path = input_path
    if os.path.exists(output_path) and input_path != output_path:
        print("delete ", output_path)
        shutil.rmtree(output_path)
    if input_path != output_path:
        shutil.copytree(input_path, output_path)
    meta_json = generate_meta_v2(output_path)
    json.dump(meta_json, open(os.path.join(output_path, "meta.json"), "w", encoding="utf8"), indent=4)
    dist_path = os.path.join(os.path.dirname(__file__), "dist")
    for filename in os.listdir(dist_path):
        path = os.path.join(dist_path, filename)
        target = os.path.join(output_path, filename)
        if os.path.exists(target):
            if os.path.isdir(target):
                shutil.rmtree(target)
            else:
                os.remove(target)
        if os.path.isdir(path):
            shutil.copytree(path, target)
        else:
            shutil.copy(path, target)


def main():
    if len(sys.argv) < 2 or sys.argv[1] == "-h":
        print(
            """
Usage: minireport data_path report_output_path\n\n
    data_path: default report data folder is the folder named 'output' in case folder\n
    report_output_path: anyplace you want
            """
        )
        exit(0)
    input_path = sys.argv[1]
    output_path = None
    if len(sys.argv) > 2:
        output_path = sys.argv[2]
    return imp_main(input_path, output_path)


if __name__ == "__main__":
    # outputs = "D:/dddd/weixin/miniumtest/miniprogram-demo-test/outputs"
    # meta_json = generate_meta_v2(outputs)
    # print(meta_json)
    main()



# -*- coding: utf-8 -*-
"""
@author: 'xiazeng'
@created: 2016/12/12
"""

import datetime
import json
import re
import logging
import subprocess
import threading
import time
import sys

from at.utils import magic
from at.core import uixml, config, resguard, basedriver
from at.core.exceptions import *
from at.core.websocketcli import WebSocketCli

if sys.version_info[0] < 3:
    from Queue import Queue, Empty
    from urllib import unquote
else:
    from queue import Queue, Empty
    from urllib.parse import unquote


RESPONSE_ERROR_JSON_PARSE_ERROR = 6
RESPONSE_ERROR_UI_OBJECT_NOT_FOUND = 5
RESPONSE_ERROR_NO_SUCH_METHOD = 3
RESPONSE_ERROR_PARAMS_UNVALIDED = 4
UNKNOW_ERROR = 7

logger = logging.getLogger()


def enqueue_output(out, queue):
    for line in iter(out.readline, b''):
        queue.put(line)
    out.close()


class JavaDriver(basedriver.JavaBaseDriver):
    ACTION_QUIT = "quit"
    ACTION_GET_TIMESTAMP = "timestamp"
    ACTION_PING = "ping"
    ACTION_IMPORTPATH = "importPath"
    ACTION_GETLIST = "getrespondlist"
    ACTION_ONOFF = "on_off"
    ACTION_HAS_READY = "hasReady"
    ACTION_BASEUI = "baseUi"
    ACTION_UI_DEVICE = "uiDevice"
    ACTION_UI_CFG = "uiCfg"
    ACTION_PY_CFG = "pyCfg"
    ACTION_HTTP_GET = "httpGet"
    ACTION_UI_SELECTOR = "pySelector"
    ACTION_LOGCAT = "logcat"
    ACTION_UPLOAD = "upload"
    ACTION_CONTEXT_UTIL = "contextUtil"
    ACTION_SYS_HANDLER = "sysDialogHandler"
    ACTION_DIALOG_HANDLER = "appDialogHandler"
    ACTION_AT_DEVICE = "aTDevice"
    ACTION_SCREEN_CAPTURE = "ScreenCapture"

    UPLOAD_DIR = "/data/local/tmp"
    SERVER_PORT = 9999
    WEBSOCKET_SERVER_PORT = 9998

    java_drivers = {}

    def __init__(self, serial, uiautomator=config.UIAUTOMATOR2, open_atstub=True):
        super(JavaDriver, self).__init__(serial)
        self._server_thread = None
        self.mobile_gap_ms = 0  # 跟手机的时机差距，毫秒
        self.app_outputs = []
        self.device_operation_records = []
        self.operation_list = []
        self.ui_trace_list = []
        self._capture_op = True
        self._server_cmd = ""
        self._stop_cmd = ""
        self.uiautomator_version = uiautomator
        self._last_op_msg = None
        self._run_app_server_retry = 1
        self._init(open_atstub=open_atstub)
        self._remove_kinda_main = False

    def release_variable(self):
        self.device_operation_records = []
        self.operation_list = []
        self.ui_trace_list = []

    def reconnect(self):
        self.close_remote()
        time.sleep(1)
        self._init()

    def _init(self, open_atstub=False):
        self._run_app_server_retry = 2
        cmd = "uiautomator runtest %s -c %s#%s " % (config.JAR_STUB_FILENAME,
                                                    config.JAR_STUB_CLASS,
                                                    config.STUB_CASE_NAME)
        install_ret = self._init_uiautomator2()
        if install_ret:
            if magic.is_windows():
                cmd = 'am instrument -w -r  -e class "%s#%s" "%s/androidx.test.runner.AndroidJUnitRunner"' % (
                    config.TEST_APP_CLS, config.STUB_CASE_NAME, config.TEST_APP_PKG
                )
            else:
                cmd = "am instrument -w -r  -e class '%s#%s' '%s/androidx.test.runner.AndroidJUnitRunner'" % (
                    config.TEST_APP_CLS, config.STUB_CASE_NAME, config.TEST_APP_PKG
                )
            self._stop_cmd = "am instrument -w -r  -e class '%s#null' '%s/androidx.test.runner.AndroidJUnitRunner'" % (
                config.TEST_APP_CLS, config.TEST_APP_PKG
            )
        else:
            raise RuntimeError("init uiautomator2 failed1")

        self.adb.run_shell(f"dumpsys deviceidle  whitelist +{config.TEST_APP_PKG}")
        if self.adb.sdk_version >= 24:
            self.adb.run_shell("cmd appops  set com.tencent.weautomator RUN_IN_BACKGROUND allow")

        cmd = self.adb.prefix() + " shell " + cmd
        self._server_cmd = cmd
        self.adb.forward(self._port, JavaDriver.SERVER_PORT)
        if not self.ping():
            self.set_app_server_run(False)
            self.run_remote_server()
            if open_atstub:
                if not self.wait_for_ui_ready(5):
                    self.adb.start_app(config.TEST_APP_PKG, config.TEST_APK_ACT)
                    time.sleep(2)
                    self.adb.press_back()
            if not self.wait_for_ui_ready():
                self.set_app_server_run(False)
                raise FailedConnectAtServer("启动AtServer失败")
        else:
            self.set_app_server_run(True)
        self.adb.forward(JavaDriver.WEBSOCKET_SERVER_PORT, JavaDriver.WEBSOCKET_SERVER_PORT)
        self.websocket = WebSocketCli(JavaDriver.WEBSOCKET_SERVER_PORT)
        for action, http_params, kwargs in self.requests_when_reconnect:
            self.do_request(action, http_params, **kwargs)

    def _init_uiautomator2(self):
        ret = True
        if not self.adb.pkg_has_installed(config.TEST_APP_PKG):
            ret = self.adb.install(config.STUB_APK_PATH, opt="-t -r")
            if not ret:
                return False
        return ret

    def _init_uiautomator(self):
        ret = False
        if not self.adb.pkg_has_installed(config.TEST_APP_PKG):
            ret = self.adb.install(config.STUB_APK_PATH)
            if not ret:
                return False
        return ret

    @classmethod
    def apply_driver(cls, serial, version=config.UIAUTOMATOR2, open_atstub=True):
        assert serial is not None
        if serial in cls.java_drivers:
            # logger.debug("use cache JavaDriver, %s, %d", str(serial), id(cls.java_drivers[serial]))
            return cls.java_drivers[serial]
        else:
            jd = JavaDriver(serial, version, open_atstub)
            logger.info("create JavaDriver, %s, %d", str(serial), id(jd))
            cls.java_drivers[serial] = jd
            return jd

    @classmethod
    def release_driver(cls, serial):
        logger.info(serial)
        if serial in cls.java_drivers:
            jd = cls.java_drivers[serial]
            del cls.java_drivers[serial]
            jd.close()

    def remove_kinda_main(self, is_remove=True):
        self.request_configure('setUseCustom', [is_remove])  # 屏蔽非最上层页面元素
        self.request_configure('setBlockKindaMain', [is_remove])
        self._remove_kinda_main = is_remove

    def run_remote_server(self, max_wait_timeout=15):
        if self._server_thread is None:
            s = time.time()
            try_count = 0
            while time.time() - s < max_wait_timeout:
                if not self.adb.app_is_running("uiautomator"):
                    break
                pid = self.adb.get_android_pid("uiautomator")
                self.adb.run_shell("kill %s" % pid)
                time.sleep(0.1)
                try_count += 1
                logger.error("uiautomator is running, try %d" % try_count)
            else:
                self._last_error = u"UiAutomator已经被占用"
                raise AtError(u"UiAutomator已经被占用")
            self._server_thread = threading.Thread(target=self._run_app_server, args=(self._server_cmd,))
            self._server_thread.setDaemon(True)
            self._server_thread.start()
        else:
            logger.error("last thread has not stop")

    def wait_for_ui_ready(self, timeout=30):
        s = time.time()
        logger.debug("wait_for_ui_ready start")
        t = time.time() - s
        while t < timeout and not self.is_remote_running():  # todo: 偶现adb命令阻塞卡住
            time.sleep(0.2)
            t = time.time() - s
        logger.debug(", %.2f " % t)

        if not self.is_remote_running():
            logger.debug(u"launch uiautomator timeout: %dms" % ((time.time() - s)*1000))
            return False
        if time.time() - s > timeout:
            # raise FailedConnectAtServer(u"AtServer启动失败，可能uiautomator被占用，请执行adb命令检查手机环境")
            logger.debug("init atserver timeout")
            return False
        while t < timeout:
            if self.ping():
                break
            time.sleep(0.2)
            t = time.time() - s
        if t >= timeout:
            # raise FailedConnectAtServer(u"AtServer启动失败，手机比较卡，请执行adb命令检查手机环境")
            return False
        logger.info("At init completed")
        return True

    def _run_app_server(self, cmd):
        logger.info(cmd)
        self._run_app_server_retry -= 1
        if magic.is_windows():
            process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                   shell=True)
        else:
            args = re.split(r"\s+", cmd)
            process = subprocess.Popen(args, close_fds=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE,
                                       shell=False)
        q = Queue()
        t = threading.Thread(target=enqueue_output, args=(process.stdout, q))
        t.daemon = True  # thread dies with the program
        t.start()
        t = threading.Thread(target=enqueue_output, args=(process.stderr, q))
        t.daemon = True  # thread dies with the program
        t.start()
        is_crashed = False
        s = time.time()
        while process.poll() is None or not q.empty():
            line = None
            try:
                line = q.get(timeout=0.1)
                if magic.is_windows():
                    line = line.decode("gbk")
                else:
                    line = line.decode("utf-8")
            except Empty:
                pass
            except:
                logger.exception("not catch exception")
            if line is not None and len(line.strip()) > 0 and "CDS" not in line:
                if "test=%s" % config.STUB_CASE_NAME in line:
                    # 检查到用例真正运行才算启动成功
                    if not self._app_server_is_run:
                        self._cmd_process = process
                        self.set_app_server_run(True)
                self.app_outputs.append(line)
                if "close [socket]" not in line:
                    logger.debug("java print:" + line.strip())
                if "Process crashed" in line:
                    is_crashed = True
        logger.info("subprocess stopped, ret code:" + str(process.returncode))
        self._server_thread = None
        self.set_app_server_run(False)
        return is_crashed

    def ping(self):
        try:
            s = time.time()
            result = self.no_check_request(self.ACTION_PING)
            costs = (time.time() - s) * 1000
            if result:
                if isinstance(result, dict) and 'currentTimeMillis' in result:
                    mobile_millis = result['currentTimeMillis']
                    self.mobile_gap_ms = mobile_millis - (s * 1000 + costs / 2)
                return True
            else:
                return False
        except Exception:
            logger.error("ping failed")
            return False

    def getrespondlist(self):
        ret = self.do_request(self.ACTION_GETLIST)
        return ret

    def importPath(self, str):
        ret = self.do_request(self.ACTION_IMPORTPATH, str)

    def on_off(self, on):
        ret = self.do_request(self.ACTION_ONOFF, on)

    def ui_is_ready(self):
        return self.do_request(self.ACTION_HAS_READY, None)

    def dump_ui(self, retry=3):
        """
        获取顶层窗口的views, 获取的views是按照从上到下顺序，重试5次
        todo: 重构时要放到uidevice里面去
        """
        views = []  # type: list[uixml.UiView]
        i = 0
        for i in range(retry):
            res = self.request_at_device("dumpUi", [])
            if res and len(res.strip()) != 0:
                if self._remove_kinda_main:
                    res = uixml.remove_kinda_window(res)
                views = uixml.window_dump_parse_str(res, resguard.Resguard.get_resguard(self.serial))
                for view in views:
                    if view.size != 0:
                        return views
            time.sleep(i+0.5)
        logger.debug('views len: %s, try:%s', len(views), i+1)
        return views

    def get_ui_views(self):
        """

        :rtype: list[uixml.UiView]
        """
        ui_views = []
        for us in self.dump_all_views():
            ui_views += us
        return ui_views

    def get_all_window_xmls(self):
        """
        获取所有窗口的views, 获取的views是按照从上到下顺序，重试5次
        todo: 重构时要放到uidevice里面去
        """
        for i in range(3):
            res = self.request_at_device("dumpXmls", [])
            if res and len(res) != 0:
                return res
            logger.info("dumpXmls return empty value, retry again")
            time.sleep(i+0.5)
        logger.error("get_all_window_xmls failed")
        return []

    def dump_all_views(self):
        """
        获取所有窗口的views, 获取的views是按照从上到下顺序，重试5次
        todo: 重构时要放到uidevice里面去
        """
        views_list = []  # type: list[list[uixml.UiView]]
        for views_str in self.get_all_window_xmls():
            if self._remove_kinda_main:
                views_str = uixml.remove_kinda_window(views_str)
            views = uixml.window_dump_parse_str(views_str, resguard.Resguard.get_resguard(self.serial))
            views_list.append(views)
        return views_list

    def dump_activity_proxy(self):
        for i in range(5):
            res = self.request_at_device("dumpUi", [])
            if res and len(res.strip()) != 0:
                return uixml.window_dump_2_activity_proxy(res, resguard.Resguard.get_resguard(self.serial))
            time.sleep(i+1)
        return None

    def network_is_ok(self, url="www.qq.com"):
        return self.request_java(self.ACTION_HTTP_GET, [url])

    def has_view(self, selector):
        for v in self.dump_ui():
            if v.match(selector):
                return True
        return False

    def request_action(self, action, method, params):
        action = action + "/" + method
        return self.request_java(action, params)

    def request_ui_method(self, method, params):
        """
        send cmd to mobile, ask for a uiautormator action
        """
        action = self.ACTION_BASEUI + "/" + method
        return self.request_java(action, params)

    def request_at_device(self, method, params=None):
        if params is None:
            params = []
        return self.request_action(self.ACTION_AT_DEVICE, method, params)

    def request_screen_capture(self, method, params=None):
        if params is None:
            params = []
        action = self.ACTION_SCREEN_CAPTURE + "/" + method
        return self.request_java(action, params)

    def request_ui_device(self, method, params=None):
        if params is None:
            params = []
        action = self.ACTION_UI_DEVICE + "/" + method
        return self.request_java(action, params)

    def request_ui_configure(self, method, params=None):
        if params is None:
            params = []
        action = self.ACTION_UI_CFG + "/" + method
        return self.request_java(action, params)

    def request_configure(self, method, params=None):
        if params is None:
            params = []
        action = self.ACTION_PY_CFG + "/" + method
        return self.request_java(action, params)

    def request_logcat(self, method, params=None):
        if params is None:
            params = []
        action = self.ACTION_LOGCAT + "/" + method
        return self.request_java(action, params)

    def request_context(self, method, params=None):
        if params is None:
            params = []
        action = self.ACTION_CONTEXT_UTIL + "/" + method
        return self.request_java(action, params)

    def request_sys_handler(self, method, params=None):
        if params is None:
            params = []
        action = self.ACTION_SYS_HANDLER + "/" + method
        return self.request_java(action, params, timeout=90)

    def request_dialog_handler(self, method, params=None):
        if params is None:
            params = []
        action = self.ACTION_DIALOG_HANDLER + "/" + method
        return self.request_java(action, params)

    def request_ui_selector(self, selectors, method, params=None, child_selectors=None, parent_selector=None):
        if params is None:
            params = []
        # logger.info(method)
        action = self.ACTION_UI_SELECTOR + "/" + method
        http_params = dict()
        http_params["params"] = unquote(json.dumps(params))
        http_params["UiSelector"] = unquote(json.dumps(selectors))
        return self.do_request(action, http_params)

    def request_java(self, action, params, **kwargs):
        http_params = dict()
        logger.debug("%s params %s", action, json.dumps(params, ensure_ascii=False))
        http_params["params"] = unquote(json.dumps(params))
        return self.do_request(action, http_params, **kwargs)

    def close_remote(self):
        if self.is_remote_running():
            self.adb.run_shell(self._stop_cmd)
            if self._port:
                self.adb.forward_remove(self._port)
            self.wait_remote_finish()

    def get_unix_ts(self):
        return self.do_request(self.ACTION_GET_TIMESTAMP)

    def set_capture_op(self, true_or_false):
        """
        设置操作是否自动截图
        :param true_or_false:
        :return:
        """
        self._capture_op = true_or_false

    def get_capture_op(self):
        return self._capture_op

    def push_op(self, msg, *args, **kwargs):
        if not self._capture_op:
            return
        display = kwargs.get('display')
        plus_micro_seconds = time.time() - self._start_ts
        dt = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        msg = msg % args
        if u"检查" not in msg:
            if self._last_op_msg:
                self.trigger("event_capture", self._last_op_msg[:128], display)
            else:
                self.trigger("event_capture", "start", display)
            self._last_op_msg = msg
        self.device_operation_records.append(u"%s%10.2fs %s" % (dt, plus_micro_seconds, msg))
        self.operation_list.append({
            "mobile_time_ms": int(time.time() * 1000 + self.mobile_gap_ms),
            "costs": plus_micro_seconds,  # 距离最开始的时间,
            "msg": msg
        })



#!/usr/bin/env python3
# created 2021/6/4 by xiazeng
import logging
import attestbase
import at.webdriver.driver


class WebDriverTest(attestbase.AtTestBase):

    def get_h5_texts(self, process_name):
        """
        :param process_name: 测试进程支付com.tencent.mm:tools， mp业务：com.tencent.mm:toolsmp，搜一搜: com.tencent.mm
        :return:
        """
        pid = self.adb.get_android_pid(process_name)
        serial = self.at.serial

        web_sock = at.webdriver.driver.get_local_sock(serial, "xweb_devtools_remote_%d" % pid)
        if not web_sock:
            return []
        page = at.webdriver.driver.WebPage(web_sock, serial)
        texts = page.invoke_js('getTexts')
        for text in texts:
            logging.info(text)
        web_sock.close()
        return texts

    def test_dump_texts(self):
        self.get_h5_texts("com.tencent.mm:tools")
        self.get_h5_texts("com.tencent.mm")
        self.get_h5_texts("com.tencent.mm:toolsmp")


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Author:         yopofeng
Filename:       emitter.py
Create time:    2021/9/24 21:21
Description:    事件分发器

"""
from pyee import EventEmitter
from typing import Callable

class MyEventEmitter(EventEmitter):
    def remove_listener(self, event: str, f: Callable) -> None:
        """安全地移除监听函数

        :param str event: 事件名
        :param Callable f: 监听函数
        :return None: None
        """
        with self._lock:
            if f not in self._events.get(event, {}):
                return
            return self._remove_listener(event, f)
    
    def remove_class_listener(self, event: str, f: Callable) -> None:
        """移除目标类的监听函数(防止因为监听事件而无法释放类实例)

        :param str event: 事件名
        :param Callable f: 监听函数, 需要有__self__
        :return None: None
        """
        with self._lock:
            listeners = self.listeners(event)
            if not listeners:
                return
            cls = f.__self__.__class__
            for listener in listeners:
                if not getattr(listener, "__self__", None):
                    continue
                if listener.__self__.__class__ is cls and listener.__name__ == f.__name__:
                    self._remove_listener(event, listener)

# global emitter
ee = MyEventEmitter()



#!/usr/bin/env python3
# Created by xiazeng on 2020/5/27
import logging
import at.miniprogram.weapp
from attestbase import AtTestBase

logger = logging.getLogger()


class MiniTest(AtTestBase):
    def test_click_pay(self):
        m = at.miniprogram.weapp.WeApp(None, is_luggage=True)
        m.current_page.by_css_selector("body > wx-view > wx-button:nth-child(4)").click()


#!/usr/bin/env python3
import sys

__version__ = "1.4.1"

from setuptools import setup, find_packages

# We do not support Python <3.8
if sys.version_info < (3, 8):
    print(
        "Unfortunately, your python version is not supported!\n"
        "Please upgrade at least to Python 3.8!",
        file=sys.stderr,
    )
    sys.exit(1)

install_requires = [
    "ddt",
    "pyyaml",
    "websocket_client>=1.3,<1.4",  # 1.4.0后有个兼容报错
    "requests",
    "future",
    "pyee",
    "typing_extensions"
]

entry_points = {
    "console_scripts": [
        "minitest=minium.framework.loader:main",
        "minidoctor=minium.framework.minidoctor:main",
        "miniruntest=minium.framework.loader:main",
        "minireport=minium.framework.report:main",
        "mininative=minium.native.nativeapp:start_server",
    ]
}


config_path = "miniprogram/base_driver/version.json"
package_data = {
    "minium": [
        config_path,
        "framework/dist/*",
        "framework/dist/*/*",
        "native/wx_native/*/*",
        "native/qq_native/*/*",
        "native/*/*",
        "native/*/*/*/*",
        "native/*/*/*/*/*",
        "native/*/*/*/*/*/*",
        "utils/js/min/*",
        "utils/js/es5/*",
    ]
}

exclude_package_data = {"": ["*pyc", "readme.md", "build.py", "__pycache__"]}
extras_require = {
    "ios": ["tidevice==0.4.10",]
}

long_description = """
install:
```
pip3 install minium
```
if you use ios devices, you can install with
```
pip3 install minium[ios]
```

See the [document](https://minitest.weixin.qq.com/#/minium/Python/readme) for details
"""

if __name__ == "__main__":
    setup(
        name="minium",
        version=__version__,
        license="MIT",
        url="https://minitest.weixin.qq.com/#/",
        packages=find_packages(),
        description="Minium is the best MiniProgram auto test framework.",
        long_description=long_description,
        long_description_content_type="text/markdown",
        package_data=package_data,
        exclude_package_data=exclude_package_data,
        entry_points=entry_points,
        install_requires=install_requires,
        extras_require=extras_require,
        setup_requires=["setuptools"],
        python_requires=">=3.8",
        author="WeChat-Test",
        author_email="minitest@tencent.com",
        platforms=["MacOS", "Windows"],
        keywords=["minium", "WeApp", "MiniProgram", "Automation", "Test"],
    )



#!/usr/bin/env python3

"""
Author:         xiazeng
Create time:    2021/11/1
Description:    

"""
import hashlib
import inspect

# 计算trace文件的时候，框架底层的文件名
import os
import typing

g_base_lib_filenames = ['case_repair_sdk', "Operator.py", "case_repair_adapter.py"]


def get_stacks(ignore_file_lists=None):
    """

    :param ignore_file_lists: 忽略的文件列表，比如一些我们依赖
    :return:
    """
    if ignore_file_lists is None:
        ignore_file_lists = g_base_lib_filenames
    else:
        ignore_file_lists += g_base_lib_filenames
    frames = inspect.stack()
    stacks = []
    if len(frames) > 2:
        current_filename = frames[0][1]
        for frame in frames[1:]:
            file_seqs = frame[1].split(os.path.sep)
            if file_seqs[-2] == "unittest":
                break
            for filename in ignore_file_lists:
                # 包含基础酷
                if filename in frame[1]:
                    break
            else:
                if frame[1] != current_filename and os.path.join("at", "core") not in frame[1] and frame[4]:
                    # 只要函数和文件信息， 文件不能包含路径名
                    # filename, function, line number, statement
                    stack_info = {
                        "file": frame[1],
                        "func": frame[3],
                        "line": frame[2],
                        "statement": frame[4][0].strip()
                    }
                    stacks.append(stack_info)
    return stacks


def get_statement():
    """拿执行操作的语句"""
    return get_stacks()[0]['statement']


def get_trace_data(enable_trace_line: bool, trace_cache: dict) -> typing.Tuple[str, dict]:
    """
    获取步骤的trace_id和trace
    """
    stack_lines = []
    stacks = get_stacks()
    for stack_info in stacks:
        # 只要函数和文件信息， 文件不包含路径名
        if enable_trace_line:
            stack_lines.append("%s#%s:%d %s" %
                               (os.path.basename(stack_info['file']), stack_info['func'], stack_info['line'],
                                stack_info['statement']))
        else:
            stack_lines.append("%s#%s %s" %
                               (os.path.basename(stack_info['file']), stack_info['func'], stack_info['statement']))

    stack_lines.reverse()
    stack_str = '\n'.join(stack_lines)
    if stack_str in trace_cache:
        # 遇到循环，堆栈是一样的，这里加个识别
        trace_cache[stack_str] = trace_cache[stack_str] + 1
        stack_str += "\nindex:%s" % trace_cache[stack_str]
    else:
        trace_cache[stack_str] = 0

    _md5 = hashlib.md5()
    _md5.update(stack_str.encode("utf-8"))
    return _md5.hexdigest(), stacks[0]



# -*- coding: utf-8 -*-
import re

__author__ = 'xiazeng'

import logging
import os

logger = logging.getLogger()


class Resguard(object):
    instances = {}
    replace_pkg_func = None
    pkg = "com.tencent.mm"

    def __init__(self, filename=None):
        self._res_mapping = None

        self._id_mapping = {}
        self._path_mapping = None
        self.has_mapping_lines = False
        
        if filename is not None and not os.path.exists(filename):
            logger.warning("not exists")
        if filename is not None and os.path.exists(filename):
            mapping = None
            self.has_mapping_lines = True
            for line in open(filename, "r"):
                if not line.startswith(" ") and ":" in line:
                    if line.strip() == "res path mapping:":
                        mapping = self._path_mapping
                    elif line.strip() == "res id mapping:":
                        mapping = self._res_mapping
                if "->" in line:
                    ls = line.strip().split("->")
                    source = ls[0].strip()
                    mapping_pkg = self.pkg
                    if self.replace_pkg_func:
                        mapping_pkg = Resguard.replace_pkg_func(self.pkg)
                    if source.startswith("%s.R.id" % mapping_pkg):
                        k = source.split(".")[-1]
                        v = ls[1].strip().split(".")[-1]
                        self._id_mapping[k] = v
                    else:
                        if mapping is not None:
                            mapping[source] = ls[1].strip()

    @classmethod
    def get_resguard(cls, serial):
        assert serial is not None
        if serial in cls.instances:
            # logger.debug("use cache Resguard, %s, %d", str(serial), id(cls.instances[serial]))
            return cls.instances[serial]    # type: Resguard
        else:
            return Resguard()

    @classmethod
    def init_resguard(cls, serial, filename):
        res = Resguard(filename)
        logger.debug('init resguard: %s', filename)
        cls.instances[serial] = res

    def retrace_id(self, s):
        if s.startswith("android:"):
            return s
        if "/" in s:
            ss = s.split("/")
            for k, v in self._id_mapping.items():
                if v == ss[-1]:
                    # logger.info("%s->%s", s, k)
                    return ss[0] + "/" + k
        else:
            for k, v in self._id_mapping.items():
                if v == s:
                    logger.info("%s->%s", s, k)
                    return k
        return s

    def retrace_xml(self, xml):
        ret_xml = xml
        rid_list = re.findall(r'resource-id="(.*?)"', xml)
        for rid in set(rid_list):
            if rid:
                ret_xml = ret_xml.replace(rid, self.retrace_id(rid))
        return ret_xml

    def resgurad_id(self, s):
        if s.startswith("android:"):
            return s
        if "/" in s:
            ss = s.split("/")
            target_s = self._id_mapping[ss[-1]] if ss[-1] in self._id_mapping else ss[-1]
            if Resguard.replace_pkg_func:
                pkg = Resguard.replace_pkg_func(ss[0].split(":")[0]) + ":id"
            else:
                pkg = ss[0]
            return pkg + "/" + target_s
        else:
            return self._id_mapping[s] if s in self._id_mapping else s


if __name__ == '__main__':
    r = Resguard("/Users/mmtest/jenkins/workspace/perf_test/base/out/resguard-mapping.txt")
    print(r.retrace_id("com.tencent.mm:id/bfm"))
    print(r.resgurad_id("chatting_avatar_iv"))



# -*- coding: utf-8 -*-
# Created by xiazeng on 2020/3/10
import time
import logging
from at.tests.attestbase import AtTestBase
from at.core.stf.minitouch import MiniTouch

logger = logging.getLogger()


class MiniTouchTest(AtTestBase):
    def setUp(self):
        super(MiniTouchTest, self).setUp()
        self.front = MiniTouch(self.at.serial, input_event="/dev/input/event1")
        self.back = MiniTouch(self.at.serial, input_event="/dev/input/event0")

    def tearDown(self):
        self.front.release()
        self.back.release()

    def test_feed_back(self):
        self.front.click(399, 650)

    def test_click(self):
        self.back.click(100, 800)
        time.sleep(5)
        self.front.click(400, 1100)
        time.sleep(1)
        for i in range(11):
            self.front.click(750, 20)
            time.sleep(0.1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by xiazeng on 2018/9/4
import time
import logging

logger = logging.getLogger()

WEB_DEBUG_PORT_REGEX_MAPPING = {
    "x5": [r"webview_devtools_remote_(?P<pid>\d+)"],
    "xweb": [r"com\.tencent\.mm_devtools_remote(?P<pid>\d+)", r"xweb_devtools_remote_(?P<pid>\d+)", r"webview_devtools_remote_(?P<pid>\d+)"],
    # xweb成功伪装成系统内核
    "webkit": [r"xweb_devtools_remote_(?P<pid>\d+)", r"webview_devtools_remote_(?P<pid>\d+)"],
}

MM_TOOL_PROCESS_NAME = "com.tencent.mm:tools"

X5_CLASS_NAMES = [
    "com.tencent.tbs.core.webkit.WebView",
    "com.tencent.smtt.webkit.WebView",
    "com.tencent.tbs.core.webkit.WebView",
    "com.tencent.tbs.core.webkit.tencent.TencentWebViewProxy"
]


def get_web_type(jd, timeout=10):
    t = time.time()
    while time.time() - t < timeout:
        ui_views = jd.dump_ui()
        # print ui_views
        if contains_x5(ui_views):
            return 'x5'
        for ui_view in ui_views:
            if ui_view.rid == "com.tencent.mm:id/logo_wv_container":
                return 'xweb'
            if ui_view.cls_name == "android":
                return "webkit"
    return "xweb"   # 我们测试大部分都用web


def contains_x5(ui_views):
    for ui_view in ui_views:
        for x5_cls_name in X5_CLASS_NAMES:
            if x5_cls_name in ui_view.cls_name:
                logger.info(u"contains x5 view, %s", ui_view.cls_name)
                return True
        # if u"X5内核提供技术支持" in ui_view.text:
        #     logger.info(u"contains x5 desc, %s", ui_view)
        #     return True
    return False
